/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 * photo/image mosaic plug-in (C) 1999 Peter Kirchgessner
 * e-mail: peter@kirchgessner.net, WWW: http://www.kirchgessner.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * This plug-in generates a photo/image mosaic from a given image.
 * I.e. a composition of several images. The composition then
 * looks like the original image.
 */

#define VERSIO "1.08"
/* Event history:
 * V 1.00, PK, 20-Feb-99, Creation
 * V 1.01, PK, 24-Mar-99, Save archive images as PPM for quality = 100
 *                        Use better (but slower) scale algorithm
 *             21-Apr-99, make it work with GIMP 1.0.4
 * V 1.02, PK, 18-Jul-99, mirror images, random replace
 * V 1.03, PK, 31-Jul-99, use gimp-file-load
 * V 1.04, PK, 01-Jun-00, fix problem with alpha channel
 *                        (thanks to Enrique Pujol Cremades)
 *                        Internationalization/GIMP 1.2-ready
 *                        Rename Photo Mosaic (tm) to Image Mosaic
 *                        (thanks to Aaron Sherman)
 * V 1.05, PK, 26-Sep-00, GIMP 1.1.26 ready (textual changes)
 * V 1.06, PK, 02-Oct-00, Win32 port
 * V 1.07, PK, 27-Dec-00, Dont use stdplugins-intl.h
 *                        Check for reusing images
 * V 1.08, PK, 23-Jan-01, Fix problem with reading non JPEG/PPM-files
 *                        on Win32
 */
static char ident[] = "@(#) GIMP ImageMosaic plug-in v1.08 23-Jan-01";

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "gtk/gtk.h"
#include "libgimp/gimp.h"
#include <libgimp/gimpui.h>

#ifdef PMOSAIC_IS_STD_PLUG_IN
#include "libgimp/stdplugins-intl.h"
#else
#include "libgimp/gimpintl.h"

#ifdef HAVE_LC_MESSAGES
#define INIT_I18N()     G_STMT_START{          \
  setlocale(LC_MESSAGES, "");                  \
  bindtextdomain("gimp-libgimp", LOCALEDIR);   \
  bindtextdomain("gimp-pmosaic", LOCALEDIR);   \
  textdomain("gimp-pmosaic");                  \
                        }G_STMT_END
#else
#define INIT_I18N()     G_STMT_START{          \
  bindtextdomain("gimp-libgimp", LOCALEDIR);   \
  bindtextdomain("gimp-pmosaic", LOCALEDIR);   \
  textdomain("gimp-pmosaic");                  \
                        }G_STMT_END
#endif

#define INIT_I18N_UI()  G_STMT_START{   \
  gtk_set_locale();                     \
  setlocale (LC_NUMERIC, "C");          \
  INIT_I18N();                          \
                        }G_STMT_END

#endif

#include "pmsc.h"

/* Declare local functions
 */
static void      query  (void);
static void      run    (char      *name,
			 int        nparams,
			 GimpParam *param,
			 int       *nreturn_vals,
			 GimpParam **return_vals);


/* Define how the plug-in works. */
typedef struct {
  gint32 drawable_ID_src;          /* Drawable ID */
  int dst_width, dst_height;       /* Output size */
  int ntiles;                      /* Number of tiles */
  int keep_aspect;                 /* Keep aspect ratio */
  int exclude_grey;                /* Exclude grey images */
  int enable_v_mirror;             /* Enable vertical mirroring */
  int enable_random;               /* Enable random select tile sequence */
  int reuse_image;                 /* Reuse images */
  char mia_name[512];              /* Archive name */
} MosaicCreateVals;


typedef struct {
  char src_dir[512];               /* Source directory */
  char dst_dir[512];               /* Destination directory */
  int maxsize;                     /* Size of thumbnails */
  int quality;                     /* JPEG-quality */
  int jpgppm_only;                 /* Read JPEG/PPM only */
} MiniCreateVals;


typedef struct {
  char mia_name[512];              /* Archive name */
} MiaCreateVals;



typedef struct {
  MosaicCreateVals create_vals;
  MiaCreateVals miacreate_vals;
  MiniCreateVals minicreate_vals;
} DialogVals;


/* Data to use for the dialog */
typedef struct
{
  GtkWidget *dialog;
  GtkWidget *mc_entry[4];  /* Mosaic Create entries */
  GtkWidget *mc_fsel;      /* Mosaic Create File select */
  int mc_keep_aspect_val;  /* Value updated by toggle */
  int mc_ex_grey_val;      /* Value updated by toggle */
  int mc_v_mirror;         /* Value updated by toggle */
  int mc_random;           /* Value updated by toggle */
  int mc_reuse;            /* Value updated by toggle */

  MosaicCreateVals *create_vals;

  GtkWidget *mia_entry[1]; /* Mosaic Archive create entry */
  GtkWidget *mia_fsel;     /* Mosaic Archive File select */

  MiaCreateVals *miacreate_vals;

  GtkWidget *mini_entry[2]; /* Mini Images create entries */
  GtkWidget *mini_fsel[2];  /* Mini Images create File select */
  int mini_scale_value[2];  /* maxsize/quality updated by scales */
  int mini_jpgppm_only;     /* Value updated by toggle */

  MiniCreateVals *minicreate_vals;

  gint run;
} MosaicInterface;

/* Image structure used with mscic-routines */
typedef struct
{
  gint32 drawable_ID;
  GimpDrawable *drawable;
} GMOSAIC_IMAGE;


static gint32    create_new_image (char *filename, guint width, guint height,
                   GimpImageType gdtype, gint32 *layer_ID,
                   GimpDrawable **drawable, GimpPixelRgn *pixel_rgn);

static gint32    mosaic_create (const MosaicCreateVals *crvals);
static int       mosaic_gimage_to_mimage (gint32 image_ID, MOSAIC_IMAGE *mi);
static void      mosaic_indexed_to_rgb (unsigned char *dst,
                   const unsigned char *src, int width, int height,
                   const unsigned char *ctab);
static void      mosaic_indexeda_to_rgb (unsigned char *dst,
                   const unsigned char *src, int width, int height,
                   const unsigned char *ctab, int bpp);
static void      mosaic_gray_to_rgb (unsigned char *dst,
                   const unsigned char *src, int width, int height);
static void      mosaic_graya_to_rgb (unsigned char *dst,
                   const unsigned char *src, int width, int height, int bpp);
static void      mosaic_rgba_to_rgb (unsigned char *dst,
                   const unsigned char *src, int width, int height, int bpp);
static int       mosaic_drawable_to_miimage (gint32 drawable_ID,
                   MOSAIC_IMAGE *mi);
static int       mosaic_imgfile_read (const char *filename, MOSAIC_IMAGE *mi);
static gint32    mosaic_mini_create (const MiniCreateVals *minicreate_vals);
static gint32    mosaic_mia_create (const MiaCreateVals *miacreate_vals);

static int       check_mosaiccreate_vals (MosaicCreateVals *crvals);
static int       check_minicreate_vals (MiniCreateVals *crvals);
static int       check_miacreate_vals (MiaCreateVals *crvals);

static void      mosaic_add_browse_button (GtkWidget *vbox,
                     void (*callback)(GtkWidget *widget, gpointer data),
                     gpointer data);
static void      mosaic_add_create_button (GtkWidget *vbox,
                     void (*callback)(GtkWidget *widget, gpointer data),
                     gpointer data);
static void      mosaic_add_label_entries (GtkWidget *vbox, int n_prop,
                     char *label_text[], GtkWidget **entry, char *entry_text[]);
static void      mosaic_add_label_scales (GtkWidget *vbox, int n_prop,
                     char *label_text[], int *range_value,
                     int *actual_value, int *updated_scale_val);
static void      mosaic_add_label_toggle (GtkWidget *box,
                     const char *label_text, int start_val,
                     int *updated_toggle_val);

static gint      mosaic_dialog         (DialogVals *dvals);
static void      mosaic_destroy_datawidget (GtkWidget *widget, gpointer data);
static void      mosaic_toggle_update  (GtkWidget *widget, gpointer data);
static void      mosaic_create_callback (GtkWidget *widget, gpointer data);
static void      mosaic_quit_callback (GtkWidget *widget, gpointer data);
static void      mosaic_destroy_callback (GtkWidget *widget, gpointer data);

static void      mosaic_help_callback (GtkWidget *widget, gpointer data);
static void      mosaic_make_help (GtkWidget **g_dialog);

static GtkWidget *mosaic_make_fsel (GtkWidget **dialog, char *title,
                         char *init_filename, int hide_fileops,
                         void (*browse_ok_callback)(GtkWidget *, gpointer),
                         gpointer browsedata);

static void      mosaic_iscale_update (GtkAdjustment *adjustment,
                                       int *scale_val);

static void      mosaic_browse_callback (GtkWidget *widget, gpointer data);
static void      mosaic_browse_ok_callback (GtkWidget *widget, gpointer data);

static void      mia_create_callback (GtkWidget *widget, gpointer data);
static void      mia_browse_callback (GtkWidget *widget, gpointer data);
static void      mia_browse_ok_callback (GtkWidget *widget, gpointer data);

static void      mini_create_callback (GtkWidget *widget, gpointer data);
static void      mini_browse_src_callback (GtkWidget *widget, gpointer data);
static void      mini_browse_src_ok_callback (GtkWidget *widget, gpointer data);
static void      mini_browse_dst_callback (GtkWidget *widget, gpointer data);
static void      mini_browse_dst_ok_callback (GtkWidget *widget, gpointer data);


GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};

static MosaicCreateVals init_mosaiccreate_vals =
{
  -1,          /* Drawable ID source */
  512, 512,    /* Use size from archive images */
  256,         /* Number of tiles */
  1,           /* Keep aspect ratio */
  1,           /* Exclude grey */
  1,           /* Enable vertical mirroring */
  1,           /* Enable random sequence for tile replacement */
  1,           /* Reuse images */
  ""           /* Archive name not specified */
};

static MiaCreateVals init_miacreate_vals =
{
  ""           /* Archive name not specified */
};

static MiniCreateVals init_minicreate_vals =
{
  "/cdrom",    /* Start with cdrom */
  "",          /* Archive name not specified */
  128,         /* Maximum size of thumbnails */
  97,          /* JPEG-quality */
  1            /* Read JPG/PPM only */
};


static GimpRunModeType run_mode;

static void
mosaic_progress_update (double val)

{
 gimp_progress_update (val);
}


MAIN ()

static void
query ()
{
  static GimpParamDef mosaicdialog_args[] =
  {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE, "image", "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" }
  };
  static GimpParamDef mosaicdialog_return_vals[] =
  {
    { GIMP_PDB_IMAGE, "image", "(not used)" }
  };

  static GimpParamDef mosaiccreate_args[] =
  {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE, "image", "Input image (notused)" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
    { GIMP_PDB_INT32, "destination_width", "Width of destination image\
 (0: use size given by archive)" },
    { GIMP_PDB_INT32, "destination_height", "Height of destination image\
 (0: use size given by archive)" },
    { GIMP_PDB_INT32, "n_tiles", "number of tiles" },
    { GIMP_PDB_INT32, "num_entries", "Number of entries in next array" },
    { GIMP_PDB_INT32ARRAY, "data_array", "[0]: flag 'keep aspect ratio' (0/1),\
 [1]: flag 'exclude grey images' (0/1),\
 [2]: flag 'allow vertical mirroring' (0/1),\
 [3]: flag 'allow random tile sequence' (0/1),\
 [4]: flag 'allow reuse of images' (0/1)" },
    { GIMP_PDB_STRING, "mia_name", "Path to Mosaic Archive"}
  };
  static GimpParamDef mosaiccreate_return_vals[] =
  {
    { GIMP_PDB_IMAGE, "new_image", "Output image" }
  };

  static GimpParamDef minicreate_args[] =
  {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_STRING, "srcdir", "Source directory" },
    { GIMP_PDB_STRING, "dstdir", "Destination directory" },
    { GIMP_PDB_INT32, "num_entries", "Number of entries in next array" },
    { GIMP_PDB_INT32ARRAY, "data_array", "[0]: maximum size of thumbnails\
 [1]: jpeg-quality, [2]: flag read jpg/ppm only"}
  };

  static GimpParamDef miacreate_args[] =
  {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_STRING, "mia_name", "Path to Mosaic Archive" }
  };
  static int mosaicdialog_nargs = sizeof (mosaicdialog_args)
                          / sizeof (mosaicdialog_args[0]);
  static int mosaicdialog_nreturn_vals = sizeof (mosaicdialog_return_vals)
                          / sizeof (mosaicdialog_return_vals[0]);
  static int mosaiccreate_nargs = sizeof (mosaiccreate_args)
                          / sizeof (mosaiccreate_args[0]);
  static int mosaiccreate_nreturn_vals = sizeof (mosaiccreate_return_vals)
                          / sizeof (mosaiccreate_return_vals[0]);
  static int minicreate_nargs = sizeof (minicreate_args)
                          / sizeof (minicreate_args[0]);
  static int minicreate_nreturn_vals = 0;
  static int miacreate_nargs = sizeof (miacreate_args)
                          / sizeof (miacreate_args[0]);
  static int miacreate_nreturn_vals = 0;

#ifndef PMOSAIC_IS_STD_PLUG_IN
  gimp_plugin_domain_register ("gimp-pmosaic", NULL);
#endif
  gimp_install_procedure ("plug_in_pmosaic_dialog",
			  "Handle image mosaic dialog",
			  "Handle image mosaic dialog",
			  "Peter Kirchgessner",
			  "Peter Kirchgessner",
			  "1999",
			  N_("<Image>/Filters/Render/Image Mosaic..."),
			  "RGB*",
			  GIMP_PLUGIN,
			  mosaicdialog_nargs, mosaicdialog_nreturn_vals,
			  mosaicdialog_args, mosaicdialog_return_vals);

  gimp_install_procedure ("plug_in_pmosaic_create",
                          "Create mosaic of images from input image and \
mosaic archive",
                          "Create mosaic of images from input image and \
mosaic archive",
                          "Peter Kirchgessner",
                          "Peter Kirchgessner",
                          "1999",
                          NULL,
                          "RGB*",
                          GIMP_PLUGIN,
                          mosaiccreate_nargs, mosaiccreate_nreturn_vals,
                          mosaiccreate_args, mosaiccreate_return_vals);

  gimp_install_procedure ("plug_in_pmosaic_mini_create",
			  "Create images for mosaic archive",
			  "Create images for mosaic archive",
			  "Peter Kirchgessner",
			  "Peter Kirchgessner",
			  "1999",
			  NULL,
			  NULL,
			  GIMP_PLUGIN,
			  minicreate_nargs, minicreate_nreturn_vals,
			  minicreate_args, NULL);

  gimp_install_procedure ("plug_in_pmosaic_mia_create",
			  "Create a mosaic archive",
			  "Create a mosaic archive",
			  "Peter Kirchgessner",
			  "Peter Kirchgessner",
			  "1999",
			  NULL,
			  NULL,
			  GIMP_PLUGIN,
			  miacreate_nargs, miacreate_nreturn_vals,
			  miacreate_args, NULL);
}

static void
run (char    *name,
     int      nparams,
     GimpParam  *param,
     int     *nreturn_vals,
     GimpParam **return_vals)
{
  static GimpParam values[2];
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;
  gint32 image_ID;
  int k;

  INIT_I18N_UI();

  run_mode = param[0].data.d_int32;

  /* Only the dialog to run ? */
  if (strcmp (name, "plug_in_pmosaic_dialog") == 0)
  {DialogVals dialog_vals;

    memcpy (&(dialog_vals.create_vals), &init_mosaiccreate_vals,
            sizeof (dialog_vals.create_vals));
    memcpy (&(dialog_vals.miacreate_vals), &init_miacreate_vals,
            sizeof (dialog_vals.miacreate_vals));
    memcpy (&(dialog_vals.minicreate_vals), &init_minicreate_vals,
            sizeof (dialog_vals.minicreate_vals));

    *nreturn_vals = 2;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;
    values[1].type = GIMP_PDB_IMAGE;
    values[1].data.d_int32 = -1;   /* We always return no image */

    switch (run_mode)
    {
      case GIMP_RUN_INTERACTIVE:
      case GIMP_RUN_WITH_LAST_VALS:
        /*  Possibly retrieve data */
        gimp_get_data (name, &dialog_vals);

        /* The procedure(s) will be run from within the dialog. */
        /* If the dialog finishes, we are also finished. */
        dialog_vals.create_vals.drawable_ID_src = param[2].data.d_int32;
        mosaic_dialog (&dialog_vals);

        gimp_set_data (name, &dialog_vals, sizeof (dialog_vals));
        break;

      case GIMP_RUN_NONINTERACTIVE:   /* No non-interactive mode */
        status = GIMP_PDB_CALLING_ERROR;
        break;

      default:
        break;
    }
    values[0].data.d_status = status;
  }
  else if (strcmp (name, "plug_in_pmosaic_create") == 0)
  {MosaicCreateVals create_vals;

    memcpy (&create_vals, &init_mosaiccreate_vals, sizeof (create_vals));

    *nreturn_vals = 2;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;
    values[1].type = GIMP_PDB_IMAGE;
    values[1].data.d_int32 = -1;

    switch (run_mode)
    {
      case GIMP_RUN_INTERACTIVE: /* This only means to display progress bar */
      case GIMP_RUN_NONINTERACTIVE:
        /*  Make sure all the arguments are there!  */
        if (nparams != 9)
          status = GIMP_PDB_CALLING_ERROR;
        if (status == GIMP_PDB_SUCCESS)
        {
            create_vals.drawable_ID_src = param[2].data.d_int32;
            create_vals.dst_width = param[3].data.d_int32;
            create_vals.dst_height = param[4].data.d_int32;
            create_vals.ntiles = param[5].data.d_int32;
            create_vals.keep_aspect = (param[6].data.d_int32 >= 1) ?
                   param[7].data.d_int32array[0] : 1;
            create_vals.exclude_grey = (param[6].data.d_int32 >= 2) ?
                   param[7].data.d_int32array[1] : 1;
            create_vals.enable_v_mirror = (param[6].data.d_int32 >= 3) ?
                   param[7].data.d_int32array[2] : 1;
            create_vals.enable_random = (param[6].data.d_int32 >= 4) ?
                   param[7].data.d_int32array[3] : 1;
            create_vals.reuse_image = (param[6].data.d_int32 >= 5) ?
                   param[7].data.d_int32array[4] : 1;
            k = sizeof (create_vals.mia_name);
            strncpy (create_vals.mia_name, param[8].data.d_string, k);
            create_vals.mia_name[k-1] = '\0';
        }
        break;

      case GIMP_RUN_WITH_LAST_VALS:
        /*  Possibly retrieve data  */
        gimp_get_data ("plug_in_pmosaic_create", &create_vals);
        break;

      default:
        break;
    }
    if (   (status == GIMP_PDB_SUCCESS)
        && (check_mosaiccreate_vals (&create_vals) < 0))
      status = GIMP_PDB_CALLING_ERROR;

    if (status == GIMP_PDB_SUCCESS)
    {
      if (run_mode != GIMP_RUN_NONINTERACTIVE)
      {
        gimp_progress_init (_("Composing Images..."));
        msc_progress_function = mosaic_progress_update;
        msc_message_function = (MOSAIC_MSGFUN)gimp_message;
      }
      else
      {
        msc_progress_function = NULL;
        msc_message_function = NULL;
      }

      image_ID = mosaic_create (&create_vals);

      msc_progress_function = NULL;
      msc_message_function = NULL;

      if (image_ID <= 0)
      {
        status = GIMP_PDB_EXECUTION_ERROR;
      }
      else
      {
        values[1].data.d_int32 = image_ID;
        gimp_image_undo_enable (image_ID);
        gimp_image_clean_all (image_ID);
        if (run_mode != GIMP_RUN_NONINTERACTIVE)
          gimp_display_new (image_ID);
      }
    }
    values[0].data.d_status = status;
  }
  else if (strcmp (name, "plug_in_pmosaic_mini_create") == 0)
  {MiniCreateVals minicreate_vals;

    memcpy (&minicreate_vals, &init_minicreate_vals, sizeof (minicreate_vals));

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    switch (run_mode)
    {
      case GIMP_RUN_INTERACTIVE: /* This only means to display progress bar */
      case GIMP_RUN_NONINTERACTIVE:
        /*  Make sure all the arguments are there!  */
        if (nparams != 5)
          status = GIMP_PDB_CALLING_ERROR;
        if (status == GIMP_PDB_SUCCESS)
        {
          k = sizeof (minicreate_vals.src_dir);
          strncpy (minicreate_vals.src_dir, param[1].data.d_string, k);
          minicreate_vals.src_dir[k-1] = '\0';

          k = sizeof (minicreate_vals.dst_dir);
          strncpy (minicreate_vals.dst_dir, param[2].data.d_string, k);
          minicreate_vals.dst_dir[k-1] = '\0';

          minicreate_vals.maxsize = (param[3].data.d_int32 >= 1) ?
                   param[4].data.d_int32array[0] : 128;
          minicreate_vals.quality = (param[3].data.d_int32 >= 2) ?
                   param[4].data.d_int32array[1] : 97;
          minicreate_vals.jpgppm_only = (param[3].data.d_int32 >= 3) ?
                   param[4].data.d_int32array[2] : 1;
        }
        break;

      case GIMP_RUN_WITH_LAST_VALS:
        /*  Possibly retrieve data  */
        gimp_get_data ("plug_in_pmosaic_mini_create", &minicreate_vals);
        break;

      default:
        break;
    }
    if (   (status == GIMP_PDB_SUCCESS)
        && (check_minicreate_vals (&minicreate_vals) < 0))
      status = GIMP_PDB_CALLING_ERROR;

    if (status == GIMP_PDB_SUCCESS)
    {
      if (run_mode != GIMP_RUN_NONINTERACTIVE)
      {
        gimp_progress_init (_("Creating Mosaic Archive Images..."));
        msc_progress_function = mosaic_progress_update;
        msc_message_function = (MOSAIC_MSGFUN)gimp_message;
      }
      else
      {
        msc_progress_function = NULL;
        msc_message_function = NULL;
      }

      k = mosaic_mini_create (&minicreate_vals);

      msc_progress_function = NULL;
      msc_message_function = NULL;

      if (k != 0)
      {
        status = GIMP_PDB_EXECUTION_ERROR;
      }
    }
    values[0].data.d_status = status;
  }
  else if (strcmp (name, "plug_in_pmosaic_mia_create") == 0)
  {MiaCreateVals miacreate_vals;

    miacreate_vals.mia_name[0] = '\0';

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    switch (run_mode)
    {
      case GIMP_RUN_INTERACTIVE: /* This only means to display progress bar */
      case GIMP_RUN_NONINTERACTIVE:
        /*  Make sure all the arguments are there!  */
        if (nparams != 2)
          status = GIMP_PDB_CALLING_ERROR;
        if (status == GIMP_PDB_SUCCESS)
        {
          k = sizeof (miacreate_vals.mia_name);
          strncpy (miacreate_vals.mia_name, param[1].data.d_string, k);
          miacreate_vals.mia_name[k-1] = '\0';
        }
        break;

      case GIMP_RUN_WITH_LAST_VALS:
        /*  Possibly retrieve data  */
        gimp_get_data ("plug_in_pmosaic_mia_create", &miacreate_vals);
        break;

      default:
        break;
    }
    if (   (status == GIMP_PDB_SUCCESS)
        && (check_miacreate_vals (&miacreate_vals) < 0))
      status = GIMP_PDB_CALLING_ERROR;

    if (status == GIMP_PDB_SUCCESS)
    {
      if (run_mode != GIMP_RUN_NONINTERACTIVE)
      {
        gimp_progress_init (_("Creating Mosaic Archive..."));
        msc_progress_function = mosaic_progress_update;
        msc_message_function = (MOSAIC_MSGFUN)gimp_message;
      }
      else
      {
        msc_progress_function = NULL;
        msc_message_function = NULL;
      }

      k = mosaic_mia_create (&miacreate_vals);

      msc_progress_function = NULL;
      msc_message_function = NULL;

      if (k != 0)
      {
        status = GIMP_PDB_EXECUTION_ERROR;
      }
    }
    values[0].data.d_status = status;
  }
}


/* Create an image. Sets layer_ID, drawable and rgn. Returns image_ID */
static gint32
create_new_image (char *filename,
                  guint width,
                  guint height,
                  GimpImageType gdtype,
                  gint32 *layer_ID,
                  GimpDrawable **drawable,
                  GimpPixelRgn *pixel_rgn)

{
 gint32 image_ID;
 GimpImageBaseType gitype;

 if ((gdtype == GIMP_GRAY_IMAGE) || (gdtype == GIMP_GRAYA_IMAGE))
   gitype = GIMP_GRAY;
 else if ((gdtype == GIMP_INDEXED_IMAGE) || (gdtype == GIMP_INDEXEDA_IMAGE))
   gitype = GIMP_INDEXED;
 else
   gitype = GIMP_RGB;

 image_ID = gimp_image_new (width, height, gitype);
 gimp_image_set_filename (image_ID, filename);

 *layer_ID = gimp_layer_new (image_ID, "Background", width, height,
                             gdtype, 100, GIMP_NORMAL_MODE);
 gimp_image_add_layer (image_ID, *layer_ID, 0);

 if (drawable != NULL)
 {
   *drawable = gimp_drawable_get (*layer_ID);
   if (pixel_rgn != NULL)
     gimp_pixel_rgn_init (pixel_rgn, *drawable, 0, 0, (*drawable)->width,
                          (*drawable)->height, TRUE, FALSE);
 }

 return (image_ID);
}


/* Functions to handle the image class */
static void
gimg_size_get (const MOSAIC_IMAGE_CLASS *this,
               int *width, int *height)

{const GMOSAIC_IMAGE *gimg = (const GMOSAIC_IMAGE *)this->image;

 *width = (int)gimg->drawable->width;
 *height = (int)gimg->drawable->height;
}


static int
gimg_size_set (MOSAIC_IMAGE_CLASS *this,
               int width,
               int height)

{GMOSAIC_IMAGE *gimg = (GMOSAIC_IMAGE *)this->image;
 gint32 image_ID, layer_ID;
 GimpDrawable *drawable;

 image_ID = create_new_image (_("Untitled"), width, height,
                  GIMP_RGB_IMAGE, &layer_ID, &drawable, NULL);

 gimg->drawable_ID = layer_ID;
 gimg->drawable = drawable;

 return (gimg ? 0 : -1);
}


static void
gimg_crop (const MOSAIC_IMAGE_CLASS *this,
           MOSAIC_IMAGE *img_out,
           int xpos, int ypos, int width,int height)

{const GMOSAIC_IMAGE *gimg = (const GMOSAIC_IMAGE *)this->image;
 gint32 dx, dy;
 unsigned char *src, *dst;
 unsigned char *rgbin = NULL;
 GimpDrawable *drawable = gimg->drawable;
 GimpPixelRgn pixel_rgn;
 int bpp = drawable->bpp;

 /* Initialize output image */
 img_out->width = width;
 img_out->height = height;

 /* Part to crop completely outside ? */
 if ((xpos >= drawable->width) || (ypos >= drawable->height)) return;

 /* The part to crop could be partially outside the image */
 dx = width;   dy = height;

 if (xpos + dx - 1 >= drawable->width) dx = drawable->width - xpos;
 if (ypos + dy - 1 >= drawable->height) dy = drawable->height - ypos;

 if ((dx == width) && (dy == height) && (bpp == 3))
 {
   rgbin = img_out->rgb; /* Dont need a separate input buffer */
 }
 else
 {
   rgbin = (unsigned char *)g_malloc (dx*dy*bpp);
   if (rgbin == NULL) return;
 }

 gimp_pixel_rgn_init (&pixel_rgn, drawable, xpos, ypos, dx, dy, FALSE, FALSE);
 gimp_pixel_rgn_get_rect (&pixel_rgn, rgbin, xpos, ypos, dx, dy);

 /* Do we need to move the data ? */
 if (rgbin != img_out->rgb)
 {
   src = rgbin;
   dst = img_out->rgb;
   if (drawable->bpp == 3)
   {
     while (--dy > 0)
     {
       g_memmove (dst, src, dx*3);
       dst += width*3;
       src += dx*bpp;
     }
   }
   else
   {int tm;
    register unsigned char *sm, *dm;

     while (--dy > 0)
     {
       tm = dx;
       sm = src;
       dm = dst;
       while (tm-- > 0)
       {
         *(dm++) = sm[0];
         *(dm++) = sm[1];
         *(dm++) = sm[2];
         sm += drawable->bpp;
       }
       dst += width*3;
       src += dx*bpp;
     }
   }
   g_free (rgbin);
 }
}


static void
gimg_paste (MOSAIC_IMAGE_CLASS *this,
            int xpos, int ypos,
            const MOSAIC_IMAGE *img)

{GMOSAIC_IMAGE *gimg = (GMOSAIC_IMAGE *)this->image;
 GimpPixelRgn pixel_rgn;
 unsigned char *src, *dst = (unsigned char *)ident;
 int k, dx, dy;

 /* Part to paste completely outside ? */
 if ((xpos >= gimg->drawable->width) || (ypos >= gimg->drawable->height))
   return;

 dx = img->width;
 dy = img->height;
                                           /* Some parts outside ? */
 if (xpos + dx - 1 >= gimg->drawable->width)
   dx = gimg->drawable->width - xpos;
 if (ypos + dy - 1 >= gimg->drawable->height)
   dy = gimg->drawable->height - ypos;
                                          /* Rearrange image ? */
 if ((dx != img->width) || (dy != img->height))
 {
   src = img->rgb + img->width*3;
   dst = img->rgb + dx*3;
   for (k = 1; k < dy; k++)
   {
     g_memmove (dst, src, dx*3);
     src += img->width*3;
     dst += dx*3;
   }
 }

 gimp_pixel_rgn_init (&pixel_rgn, gimg->drawable, xpos, ypos, dx, dy,
                      FALSE, FALSE);
 gimp_pixel_rgn_set_rect (&pixel_rgn, img->rgb, xpos, ypos, dx, dy);
}


/* Compose images from a mosaic archive */
static gint32
mosaic_create (const MosaicCreateVals *create_vals)

{gint32 image_ID_dst = -1;
 int retval;
 MOSAIC_IMAGE_CLASS img_in, img_out;
 GMOSAIC_IMAGE gimg_in, gimg_out;

 gimg_in.drawable_ID = create_vals->drawable_ID_src;
 gimg_in.drawable = gimp_drawable_get (create_vals->drawable_ID_src);

 mscic_init (&img_in, &gimg_in, gimg_size_get, gimg_size_set,
             gimg_crop, gimg_paste);
 mscic_init (&img_out, &gimg_out, gimg_size_get, gimg_size_set,
             gimg_crop, gimg_paste);

 retval = mscic_mosaic_create (&img_in, create_vals->mia_name,
                create_vals->ntiles, create_vals->dst_width,
                create_vals->dst_height, create_vals->keep_aspect,
                create_vals->exclude_grey,
                create_vals->enable_v_mirror,
                create_vals->enable_random,
                create_vals->reuse_image,
                &img_out);

 image_ID_dst = gimp_drawable_image_id (gimg_out.drawable_ID);

 return image_ID_dst;
}


static void
mosaic_indexed_to_rgb (register unsigned char *dst,
                       register const unsigned char *src,
                       int width, int height, const unsigned char *ctab)

{register int npix = width * height;
 register const unsigned char *rgb;

 while (npix-- > 0)
 {
   rgb = ctab + (*(src++)) * 3;  /* Get entry into color table */
   *(dst++) = *(rgb++);
   *(dst++) = *(rgb++);
   *(dst++) = *(rgb++);
 }
}


static void
mosaic_indexeda_to_rgb (register unsigned char *dst,
                        register const unsigned char *src,
                        int width, int height, const unsigned char *ctab,
                        int bpp)

{register int npix = width * height;
 register const unsigned char *rgb;

 while (npix-- > 0)
 {
   rgb = ctab + (*src) * 3;  /* Get entry into color table */
   *(dst++) = *(rgb++);
   *(dst++) = *(rgb++);
   *(dst++) = *(rgb++);
   src += bpp;
 }
}


static void
mosaic_gray_to_rgb (register unsigned char *dst,
                    register const unsigned char *src,
                    int width, int height)

{register int npix = width * height;

 while (npix-- > 0)
 {
   dst[0] = dst[1] = dst[2] = *(src++);
   dst += 3;
 }
}


static void
mosaic_graya_to_rgb (register unsigned char *dst,
                     register const unsigned char *src,
                     int width, int height, int bpp)

{register int npix = width * height;

 while (npix-- > 0)
 {
   dst[0] = dst[1] = dst[2] = *src;
   dst += 3;
   src += bpp;
 }
}


static void
mosaic_rgba_to_rgb (register unsigned char *dst,
                    register const unsigned char *src,
                    int width, int height, int bpp)

{register int npix = width * height;

 while (npix-- > 0)
 {
   *(dst++) = src[0];   *(dst++) = src[1];   *(dst++) = src[2];
   src += bpp;
 }
}


static int
mosaic_drawable_to_miimage (gint32 drawable_ID, MOSAIC_IMAGE *mi)

{int width, height, tile_height;
 int i, scanlines, ncols;
 unsigned char *data, *dst, *ctab = NULL;
 GimpPixelRgn pixel_rgn;
 GimpDrawable *drawable;
 GimpImageType drawable_type;

 drawable = gimp_drawable_get (drawable_ID);
 drawable_type = gimp_drawable_type (drawable_ID);

 if (   (drawable_type == GIMP_INDEXED_IMAGE)
     || (drawable_type == GIMP_INDEXEDA_IMAGE))
   ctab = gimp_image_get_cmap (gimp_drawable_image_id (drawable_ID), &ncols);

 width = drawable->width;
 height = drawable->height;

 /* Destination is always GIMP_RGB */
 dst = (unsigned char *)g_malloc (width * height * 3);
 if (dst == NULL)
 {
   if (ctab) g_free (ctab);
   return -1;
 }

 tile_height = gimp_tile_height ();
 gimp_pixel_rgn_init (&pixel_rgn, drawable, 0, 0, width, height, FALSE, FALSE);

 /* allocate a buffer for retrieving information from the pixel region  */
 data = (unsigned char *)g_malloc (tile_height * width * drawable->bpp);
 if (data == NULL)
 {
   g_free (ctab);
   g_free (dst);
   return -1;
 }

 mi->width = width;
 mi->height = height;
 mi->rgb = dst;

 i = 0;
 while (i < height)
 {
   scanlines = (i+tile_height-1 < height) ? tile_height : (height-i);
   gimp_pixel_rgn_get_rect (&pixel_rgn, data, 0, i, width, scanlines);

   switch (drawable_type)
   {
     case GIMP_RGB_IMAGE:
       memcpy (dst, data, width * scanlines * 3);
       break;

     case GIMP_RGBA_IMAGE:
       mosaic_rgba_to_rgb (dst, data, width, scanlines, drawable->bpp);
       break;

     case GIMP_GRAY_IMAGE:
       mosaic_gray_to_rgb (dst, data, width, scanlines);
       break;

     case GIMP_GRAYA_IMAGE:
       mosaic_graya_to_rgb (dst, data, width, scanlines, drawable->bpp);
       break;

     case GIMP_INDEXED_IMAGE:
       mosaic_indexed_to_rgb (dst, data, width, scanlines, ctab);
       break;

     case GIMP_INDEXEDA_IMAGE:
       mosaic_indexeda_to_rgb (dst, data, width, scanlines, ctab,drawable->bpp);
       break;
   }
   dst += width * scanlines * 3;
   i += scanlines;
 }
 g_free (ctab);
 g_free (data);

 gimp_drawable_detach (drawable);

 return 0;
}


/* Convert GIMP image to MOSAIC image */
static int
mosaic_gimage_to_mimage (gint32 image_ID, MOSAIC_IMAGE *mi)

{gint32 *layers;
 gint nlayers, layer_ID;
 int retval = -1;

 mi->width = mi->height = 0;
 mi->rgb = NULL;

 layers = gimp_image_get_layers (image_ID, &nlayers);
 if (nlayers <= 0) return -1;

 layer_ID = *layers;
 g_free (layers);

 retval = mosaic_drawable_to_miimage (layer_ID, mi);

 return retval;
}


/* User-function to read files */
static int
mosaic_imgfile_read (const char *filename, MOSAIC_IMAGE *mi)

{GimpParam *return_vals;
 gint nreturn_vals;
 gint32 image;
 int ret = -1;

 return_vals = gimp_run_procedure ("gimp_file_load", &nreturn_vals,
                              GIMP_PDB_INT32, GIMP_RUN_NONINTERACTIVE,
                              GIMP_PDB_STRING, filename,
                              GIMP_PDB_STRING, filename,
                              GIMP_PDB_END);

 if (return_vals[0].data.d_status == FALSE)  /* Call failed ? */
 {
   ret = -1;
 }
 else   /* Got an image ? */
 {
   image = return_vals[1].data.d_int32;
   if (image >= 0)
   {
     gimp_image_undo_disable (image);
     ret = mosaic_gimage_to_mimage (image, mi);
     gimp_image_delete (image);
   }
   else
   {
     ret = -1;
   }
 }

 gimp_destroy_params (return_vals, nreturn_vals);

 return ret;
}


/* Create Mosaic Archive Images */
static gint32
mosaic_mini_create (const MiniCreateVals *minicreate_vals)

{
 return msc_images_collect (minicreate_vals->src_dir, minicreate_vals->dst_dir,
                            minicreate_vals->maxsize, minicreate_vals->quality,
                            minicreate_vals->jpgppm_only ?
                              NULL : mosaic_imgfile_read);
}


/* Create Mosaic Archive */
static gint32
mosaic_mia_create (const MiaCreateVals *miacreate_vals)

{
 return msc_mia_create (miacreate_vals->mia_name);
}


/* Check createvals. Unreasonable values are reset to a default. */
/* If this is not possible, -1 is returned. Otherwise 0 is returned. */
static int
check_mosaiccreate_vals (MosaicCreateVals *create_vals)

{
 if (create_vals->dst_width < 0) create_vals->dst_width = 0;
 if (create_vals->dst_height < 0) create_vals->dst_height = 0;
 if (create_vals->ntiles < 9) create_vals->ntiles = 9;
 create_vals->keep_aspect = (create_vals->keep_aspect != 0);
 create_vals->exclude_grey = (create_vals->exclude_grey != 0);
 create_vals->enable_v_mirror = (create_vals->enable_v_mirror != 0);
 create_vals->enable_random = (create_vals->enable_random != 0);
 create_vals->reuse_image = (create_vals->reuse_image != 0);
 return (0);
}


/* Check mini createvals. Unreasonable values are reset to a default. */
/* If this is not possible, -1 is returned. Otherwise 0 is returned. */
static int
check_minicreate_vals (MiniCreateVals *minicreate_vals)

{
 if (minicreate_vals->src_dir[0] == '\0') return -1;
 if (minicreate_vals->dst_dir[0] == '\0') return -1;
 if (minicreate_vals->maxsize < 32) minicreate_vals->maxsize = 32;
 else if (minicreate_vals->maxsize > 255) minicreate_vals->maxsize = 255;
 if (minicreate_vals->quality < 10) minicreate_vals->quality = 10;
 else if (minicreate_vals->quality > 100) minicreate_vals->quality = 100;
 if (minicreate_vals->jpgppm_only != 0) minicreate_vals->jpgppm_only = 1;
 return (0);
}


/* Check mia createvals. Unreasonable values are reset to a default. */
/* If this is not possible, -1 is returned. Otherwise 0 is returned. */
static int
check_miacreate_vals (MiaCreateVals *miacreate_vals)

{
 if (miacreate_vals->mia_name[0] == '\0') return -1;
 return (0);
}


static void
mosaic_add_browse_button (GtkWidget *vbox,
               void (*callback)(GtkWidget *widget, gpointer data),
               gpointer data)

{GtkWidget *hbbox, *button;

 /*  browse button area */
 hbbox = gtk_hbutton_box_new ();
 gtk_button_box_set_layout(GTK_BUTTON_BOX(hbbox), GTK_BUTTONBOX_END);
 gtk_box_pack_start (GTK_BOX (vbox), hbbox, FALSE, FALSE, 5);
 gtk_widget_show (hbbox);

 button = gtk_button_new_with_label (_("Browse..."));
 gtk_signal_connect (GTK_OBJECT (button), "clicked",
                     (GtkSignalFunc) callback, data);
 gtk_box_pack_start (GTK_BOX (hbbox), button, FALSE, FALSE, 5);
 gtk_widget_show (button);
}


static void
mosaic_add_create_button (GtkWidget *vbox,
               void (*callback)(GtkWidget *widget, gpointer data),
               gpointer data)

{GtkWidget *separator, *button;

 separator = gtk_hseparator_new ();
 gtk_box_pack_start (GTK_BOX (vbox), separator, FALSE, TRUE, 0);
 gtk_widget_show (separator);

 button = gtk_button_new_with_label (_("Create"));
 gtk_signal_connect (GTK_OBJECT (button), "clicked",
                     (GtkSignalFunc) callback, data);
 gtk_box_pack_start (GTK_BOX (vbox), button, TRUE, FALSE, 5);
 gtk_widget_show (button);
}


static void
mosaic_add_label_entries (GtkWidget *vbox, int n_prop, char *label_text[],
                          GtkWidget **entry, char *entry_text[])

{GtkWidget *table, *label;
 int j;

 table = gtk_table_new (n_prop, 2, FALSE);
 gtk_table_set_row_spacings (GTK_TABLE (table), 5);
 gtk_table_set_col_spacings (GTK_TABLE (table), 5);
 gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);

 /* Create labels */
 for (j = 0; j < n_prop; j++)
 {
   label = gtk_label_new (gettext (label_text[j]));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   gtk_table_attach (GTK_TABLE (table), label, 0, 1, j, j+1,
                     GTK_FILL, GTK_FILL, 0, 0);
   gtk_widget_show (label);
 }

 /* Create entries */
 for (j = 0; j < n_prop; j++)
 {
   entry[j] = gtk_entry_new ();
   gtk_widget_set_usize (entry[j], 160, 0);
   gtk_entry_set_text (GTK_ENTRY (entry[j]), entry_text[j]);
   gtk_table_attach (GTK_TABLE (table), entry[j],
                     1, 2, j, j+1,
                     GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
   gtk_widget_show (entry[j]);
 }
 gtk_widget_show (table);
}


static void
mosaic_add_label_scales (GtkWidget *vbox, int n_prop, char *label_text[],
                         int *range_value, int *actual_value,
                         int *updated_scale_value)

{GtkWidget *table, *label, *scale;
 GtkObject *adjustment;
 int j, *rv = range_value;

 table = gtk_table_new (n_prop, 2, FALSE);
 gtk_table_set_row_spacings (GTK_TABLE (table), 5);
 gtk_table_set_col_spacings (GTK_TABLE (table), 5);
 gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);

 /* Create labels */
 for (j = 0; j < n_prop; j++)
 {
   label = gtk_label_new (gettext (label_text[j]));
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   gtk_table_attach (GTK_TABLE (table), label, 0, 1, j, j+1,
                     GTK_FILL, GTK_FILL, 0, 0);
   gtk_widget_show (label);
 }

 /* Create scales */
 for (j = 0; j < n_prop; j++)
 {
   adjustment = gtk_adjustment_new ((gfloat)actual_value[j], /* value */
                                    (gfloat)rv[0],           /* lower */
                                    (gfloat)(rv[1]+1),       /* upper */
                                    1.0, 1.0, 1.0);
   rv += 2;
   updated_scale_value[j] = actual_value[j];
   scale = gtk_hscale_new (GTK_ADJUSTMENT (adjustment));
   gtk_widget_set_usize (GTK_WIDGET (scale), 128, 30);
   gtk_scale_set_digits (GTK_SCALE (scale), 0);
   gtk_scale_set_draw_value (GTK_SCALE (scale), TRUE);
   gtk_range_set_update_policy (GTK_RANGE (scale), GTK_UPDATE_DELAYED);
   gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
                      (GtkSignalFunc) mosaic_iscale_update,
                      updated_scale_value+j);
   gtk_table_attach (GTK_TABLE (table), scale,
                     1, 2, j, j+1,
                     GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
   gtk_widget_show (scale);
 }
 gtk_widget_show (table);
}


static void
mosaic_add_label_toggle (GtkWidget *box, const char *label_text,
                         int start_val, int *updated_toggle_val)

{GtkWidget *toggle;

 toggle = gtk_check_button_new_with_label (label_text);
 gtk_box_pack_start (GTK_BOX (box), toggle, TRUE, TRUE, 0);
 *updated_toggle_val = start_val;
 gtk_signal_connect (GTK_OBJECT (toggle), "toggled",
                    (GtkSignalFunc) mosaic_toggle_update, updated_toggle_val);
 gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle), *updated_toggle_val);
 gtk_widget_show (toggle);
}


static gint
mosaic_dialog (DialogVals *dialog_vals)

{gint argc;
 gchar **argv;
 char buffers[4][1024];
 char *bufarr[4];
 int data_array[16];

 GtkWidget *dlg;
 GtkWidget *notebook;
 GtkWidget *frame;
 GtkWidget *vbox;

 MosaicInterface mosaic_ui;

 MosaicCreateVals *create_vals = &(dialog_vals->create_vals);
 MiaCreateVals *miacreate_vals = &(dialog_vals->miacreate_vals);
 MiniCreateVals *minicreate_vals = &(dialog_vals->minicreate_vals);

 static char *create_label_text[] = { N_("Width:"), N_("Height:"),
                      N_("Number of tiles:"), N_("Mosaic archive:") };
 static char *miacreate_label_text[] = { N_("Mosaic archive:") };
 static char *minicreate_label_text[] = { N_("Source directory:"),
                                          N_("Destination directory:") };
 static char *minicreate_scale_text[] = { N_("Maximum size:"), N_("Quality:") };
 static int minicreate_scale_values[] = { 32, 255, 10, 100 };

 gimp_ui_init ("pmosaic", TRUE);

 bufarr[0] = buffers[0];
 bufarr[1] = buffers[1];
 bufarr[2] = buffers[2];
 bufarr[3] = buffers[3];

 memset (&mosaic_ui, 0, sizeof (mosaic_ui));
 mosaic_ui.run = FALSE;
 mosaic_ui.dialog = NULL;
 mosaic_ui.create_vals = &(dialog_vals->create_vals);
 mosaic_ui.miacreate_vals = &(dialog_vals->miacreate_vals);
 mosaic_ui.minicreate_vals = &(dialog_vals->minicreate_vals);

 argc = 1;
 argv = g_new (gchar *, 1);
 argv[0] = g_strdup (_("Image Mosaic"));

 gtk_init (&argc, &argv);
 gtk_rc_parse (gimp_gtkrc ());
 gdk_set_use_xshm (gimp_use_xshm ());

 mosaic_ui.dialog = dlg = gtk_dialog_new ();

 dlg = gimp_dialog_new (_("Image Mosaic"), "pmosaic",
                         gimp_standard_help_func, "filters/pmosaic.html",
                         GTK_WIN_POS_NONE,
                         FALSE, TRUE, FALSE,

                         _("Close"), mosaic_quit_callback, &mosaic_ui,
                         NULL, NULL, TRUE, TRUE,

                         _("Help"), mosaic_help_callback, NULL,
                         NULL, NULL, FALSE, FALSE,

                         NULL);

 gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
                     (GtkSignalFunc) mosaic_destroy_callback, &mosaic_ui);

 /* Create notebook */
 notebook = gtk_notebook_new ();
 gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_TOP);
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), notebook,FALSE,FALSE,0);
 gtk_widget_show (notebook);

 /* The frame for creation of photo/image mosaics */
 frame = gtk_frame_new (_("Create Image Mosaic"));
 gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
 gtk_container_set_border_width (GTK_CONTAINER (frame), 5);

 vbox = gtk_vbox_new (FALSE, 0);
 gtk_container_set_border_width (GTK_CONTAINER (vbox), 5);
 gtk_container_add (GTK_CONTAINER (frame), vbox);

 /* Labels and entries (width, height, ntiles, mia name */
 sprintf (bufarr[0], "%d", create_vals->dst_width);
 sprintf (bufarr[1], "%d", create_vals->dst_height);
 sprintf (bufarr[2], "%d", create_vals->ntiles);
 sprintf (bufarr[3], "%s", create_vals->mia_name);
 mosaic_add_label_entries (vbox, 4, create_label_text,
                           &(mosaic_ui.mc_entry[0]), bufarr);

 mosaic_add_browse_button (vbox, mosaic_browse_callback, &mosaic_ui);

 mosaic_add_label_toggle (vbox, _("Keep aspect ratio"),
                          create_vals->keep_aspect,
                          &mosaic_ui.mc_keep_aspect_val);
 mosaic_add_label_toggle (vbox, _("Exclude grey images"),
                          create_vals->exclude_grey,
                          &mosaic_ui.mc_ex_grey_val);

 mosaic_add_label_toggle (vbox, _("Allow vertical mirroring"),
                          create_vals->enable_v_mirror,
                          &mosaic_ui.mc_v_mirror);

 mosaic_add_label_toggle (vbox, _("Random replace sequence"),
                          create_vals->enable_random,
                          &mosaic_ui.mc_random);

 mosaic_add_label_toggle (vbox, _("Multiple use of images"),
                          create_vals->reuse_image,
                          &mosaic_ui.mc_reuse);

 mosaic_add_create_button (vbox, mosaic_create_callback, &mosaic_ui);

 gtk_widget_show (vbox);

 gtk_notebook_append_page (GTK_NOTEBOOK (notebook), frame,
                           gtk_label_new (_("Mosaic")));
 gtk_widget_show (frame);

 /* The frame for creation of photo/image mosaic archive images */
 frame = gtk_frame_new (_("Create Mosaic Archive Images"));
 gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
 gtk_container_set_border_width (GTK_CONTAINER (frame), 5);

 vbox = gtk_vbox_new (FALSE, 0);
 gtk_container_set_border_width (GTK_CONTAINER (vbox), 5);
 gtk_container_add (GTK_CONTAINER (frame), vbox);

 /* label and entries (source directory) */
 sprintf (bufarr[0], "%s", minicreate_vals->src_dir);
 mosaic_add_label_entries (vbox, 1, minicreate_label_text,
                           &(mosaic_ui.mini_entry[0]), bufarr);

 mosaic_add_browse_button (vbox, mini_browse_src_callback, &mosaic_ui);

 /* label and entries (dest. directory) */
 sprintf (bufarr[0], "%s", minicreate_vals->dst_dir);
 mosaic_add_label_entries (vbox, 1, &(minicreate_label_text[1]),
                           &(mosaic_ui.mini_entry[1]), bufarr);

 mosaic_add_browse_button (vbox, mini_browse_dst_callback, &mosaic_ui);

 data_array[0] = minicreate_vals->maxsize;
 data_array[1] = minicreate_vals->quality;
 mosaic_add_label_scales (vbox, 2, minicreate_scale_text,
                          minicreate_scale_values,
                          data_array,
                          &(mosaic_ui.mini_scale_value[0]));

 mosaic_add_label_toggle (vbox, _("Read JPEG/PPM only"),
                          minicreate_vals->jpgppm_only,
                          &(mosaic_ui.mini_jpgppm_only));

 mosaic_add_create_button (vbox, mini_create_callback, &mosaic_ui);

 gtk_widget_show (vbox);

 gtk_notebook_append_page (GTK_NOTEBOOK (notebook), frame,
                           gtk_label_new (_("Images")));
 gtk_widget_show (frame);

 /* The frame for creation of photo/image mosaic archive */
 frame = gtk_frame_new (_("Create Mosaic Archive"));
 gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
 gtk_container_set_border_width (GTK_CONTAINER (frame), 5);

 vbox = gtk_vbox_new (FALSE, 0);
 gtk_container_set_border_width (GTK_CONTAINER (vbox), 5);
 gtk_container_add (GTK_CONTAINER (frame), vbox);

 /* label and entries (mia name) */
 sprintf (bufarr[0], "%s", miacreate_vals->mia_name);
 mosaic_add_label_entries (vbox, 1, miacreate_label_text,
                           &(mosaic_ui.mia_entry[0]), bufarr);

 mosaic_add_browse_button (vbox, mia_browse_callback, &mosaic_ui);
 mosaic_add_create_button (vbox, mia_create_callback, &mosaic_ui);

 gtk_widget_show (vbox);

 gtk_notebook_append_page (GTK_NOTEBOOK (notebook), frame,
                           gtk_label_new (_("Archive")));
 gtk_widget_show (frame);

 gtk_widget_show (dlg);

 gtk_main ();

 gdk_flush ();

 return mosaic_ui.run;
}


static void
mosaic_destroy_datawidget (GtkWidget *widget,
                           gpointer data)
{
 gtk_widget_destroy (GTK_WIDGET (data));
}


static void
mosaic_create_callback (GtkWidget *widget,
                        gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;
 MosaicCreateVals *create_vals = mosaic_ui->create_vals;
 GimpParam *return_vals;
 gint32 data_array[16];
 int k, ival, nvals, nreturn_vals;

 /* Read width */
 k = sscanf (gtk_entry_get_text (GTK_ENTRY (mosaic_ui->mc_entry[0])),
             "%d", &ival);
 if (k == 1) create_vals->dst_width = ival;

 /* Read height */
 k = sscanf (gtk_entry_get_text (GTK_ENTRY (mosaic_ui->mc_entry[1])),
             "%d", &ival);
 if (k == 1) create_vals->dst_height = ival;

 /* Read number of tiles */
 k = sscanf (gtk_entry_get_text (GTK_ENTRY (mosaic_ui->mc_entry[2])),
             "%d", &ival);
 if (k == 1) create_vals->ntiles = ival;

 /* Read mosaic archive name */
 strncpy (create_vals->mia_name,
          gtk_entry_get_text (GTK_ENTRY (mosaic_ui->mc_entry[3])),
          sizeof (create_vals->mia_name));
 create_vals->mia_name[sizeof (create_vals->mia_name)-1] = '\0';
                             /* keep aspect ratio */
 create_vals->keep_aspect = mosaic_ui->mc_keep_aspect_val;
                             /* exclude grey */
 create_vals->exclude_grey = mosaic_ui->mc_ex_grey_val;
                             /* enable vertical mirroring */
 create_vals->enable_v_mirror = mosaic_ui->mc_v_mirror;
                             /* enable random replace sequence */
 create_vals->enable_random = mosaic_ui->mc_random;
                             /* enable reuse of images */
 create_vals->reuse_image = mosaic_ui->mc_reuse;

 check_mosaiccreate_vals (create_vals);

 nvals = 0;
 data_array[nvals++] = create_vals->keep_aspect;
 data_array[nvals++] = create_vals->exclude_grey;
 data_array[nvals++] = create_vals->enable_v_mirror;
 data_array[nvals++] = create_vals->enable_random;
 data_array[nvals++] = create_vals->reuse_image;

 return_vals = gimp_run_procedure ("plug_in_pmosaic_create",
                    &nreturn_vals,
                    GIMP_PDB_INT32, GIMP_RUN_INTERACTIVE,
                    GIMP_PDB_IMAGE, -1,
                    GIMP_PDB_DRAWABLE, create_vals->drawable_ID_src,
                    GIMP_PDB_INT32, create_vals->dst_width,
                    GIMP_PDB_INT32, create_vals->dst_height,
                    GIMP_PDB_INT32, create_vals->ntiles,
                    GIMP_PDB_INT32, nvals,
                    GIMP_PDB_INT32ARRAY, data_array,
                    GIMP_PDB_STRING, create_vals->mia_name,
                    GIMP_PDB_END);

 gimp_destroy_params (return_vals, nreturn_vals);
}


static void
mosaic_toggle_update (GtkWidget *widget,
                      gpointer  data)

{int *toggle_val = (int *)data;

 *toggle_val = ((GTK_TOGGLE_BUTTON (widget)->active) != 0);
}


static void
mosaic_quit_callback (GtkWidget *widget,
                      gpointer data)
{
 gtk_main_quit ();
}


static void
mosaic_destroy_callback (GtkWidget *widget,
                         gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;
 mosaic_ui->run = TRUE;
 gtk_widget_destroy (GTK_WIDGET (mosaic_ui->dialog));
}


static void
mosaic_help_callback (GtkWidget *widget,
                      gpointer data)

{static GtkWidget *dialog = NULL;

 if (!dialog)
   mosaic_make_help (&dialog);

 if (!GTK_WIDGET_VISIBLE (dialog))
   gtk_widget_show (dialog);
 else
   gtk_widget_destroy (dialog);
}


static void
mosaic_make_help (GtkWidget **g_dialog)

{GtkWidget *dialog;
 GtkWidget *button;
 GtkWidget *vsb;
 GtkWidget *text;
 GtkWidget *table;
 GtkAdjustment *vadj;
 GdkFont   *font_strong;
 GdkFont   *font_emphasis;
 GdkFont   *font;
 char buffer[80];
 static char *helptext_1 =
 N_("\n"
   "This plug-in creates an image mosaic from an input image.\n"
   "An image mosaic is a composition of several small images (photos).\n"
   "But when viewed from a distance, it shows the input image again.\n"
   "To create an image mosaic, the input image is partitioned into\n"
   "small tiles. Each tile is then replaced by a small image,\n"
   "that has similar colours. The better the colours match the tile,\n"
   "the better the image mosaic matches the input image.\n"
   "To get good image mosaics, a large number of small images\n"
   "should be available to choose from. This collection of images\n"
   "also has to cover the complete range of colours.\n\n"
   "To save computation time for choosing the best fitting\n"
   "small image, the plug-in uses a 'Mosaic Archive'. This keeps\n"
   "the colour information for all images in the archive and the\n"
   "images themself as small JPEG- or PPM-files. The colour informations\n"
   "are kept in a separate file (Mosaic Information Archive, MIA).\n"
   "The images are located in subdirectories. The MIA represents the\n"
   "archive. The archive images must not exceed a size of 255*255 pixels.\n\n"
   "To create an image mosaic with the plug-in, three steps\n"
   "are required:\n\n");

 static char *helptext_2 =
 N_("Step 1: Create collection of small images\n"
   "\tSelect the 'Images'-tab from the dialog.\n"
   "\tEnter the source directory of your hard disc or CD-ROM\n"
   "\t\twhich you want to scan for images. The plug-in will automatically\n"
   "\t\tscan all subdirectories for images.\n"
   "\tEnter the destination directory on your hard disc which you want\n"
   "\t\tto use for the Mosaic Archive. The directory must already exist.\n"
   "\t\tWith that file-select-dialog you can create directories.\n"
   "\t\tThe plug-in creates a similar directory structure below the\n"
   "\t\tdestination directory as it finds below the source directory.\n"
   "\tAdjust the maximum size of the archive images. The larger the\n"
   "\t\timages, the better the quality. But also more disk space to use.\n"
   "\t\tSince an image mosaic is usually composed of a large number of\n"
   "\t\timages, the archive images need not to be very large. A size of\n"
   "\t\t128*128 pixels should be quite enough.\n"
   "\tAdjust the quality. This specifies the compression to use for writing\n"
   "\t\tthe archive images. If a quality of 100 is given, the archive images\n"
   "\t\tare saved as PPM files without any compression.\n"
   "\tSelect 'Read JPEG/PPM only' if you only want to search for JPEG- or\n"
   "\t\tPPM-files. If this is not selected, all files recognized by GIMP\n"
   "\t\tare read.\n"
   "\tPress 'Create' to start scanning for images, scaling them to the\n"
   "\t\trequired size and writing the archive images.\n"
   "\t\tYou can run this step multiple times for different source\n"
   "\t\tdirectories and same destination directory (but beware of\n"
   "\t\toverwriting existing files).\n\n"
   "Step 2: Create Mosaic Information Archive MIA\n"
   "\tSelect the 'Archive'-tab from the dialog.\n"
   "\tEnter the name of the MIA to create. The name of the file should be\n"
   "\t\t<directory>/<somename>.mia . <directory> is a directory that\n"
   "\t\thas been used as a destination directory in step 1.\n"
   "\tPress 'Create' to start scanning <directory> for the archive\n"
   "\t\timages placed in subdirectories and creating the MIA.\n"
   "\tNow you are ready to create image mosaics.\n\n"
   "Step 3: Create Image Mosaic\n"
   "\tSelect the 'Mosaic'-tab from the dialog.\n"
   "\tChoose width and height of the output image in pixels.\n"
   "\tChoose the number of tiles to use for the composition.\n"
   "\tEnter the name of a *.mia-file created in step 2 to select\n"
   "\t\tfrom which set of images you want to create the image mosaic.\n"
   "\tSelect 'Keep aspect ratio' if you dont want to distort your\n"
   "\t\timage.\n"
   "\tSelect 'Exclude grey images' if you dont want to use archive\n"
   "\t\timages which are grey.\n"
   "\tSelect 'Allow vertical mirroring' to enable vertical flipping\n"
   "\t\tof archive images to better replace a tile of the input image.\n"
   "\tSelect 'Random replace sequence' to enable random replacement\n"
   "\t\tof tiles by archive images. This will give a little bit better\n"
   "\t\tresult on archives with a small number of archive images.\n"
   "\tSelect 'Multiple use of images' to allow an archive image to be used\n"
   "\t\tmultiple times. If this is not selected, it could happen that no\n"
   "\t\tcorresponding archive image is found and no replacement takes place.\n"
   "\tPress 'Create' to start the composition of the image mosaic.\n\n"
   "For suggestions about the plug-in, mailto:peter@kirchgessner.net\n");

 *g_dialog = dialog = gtk_dialog_new ();

 gtk_window_set_title (GTK_WINDOW (dialog), _("Image Mosaic Help"));
 gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
 gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
                     (GtkSignalFunc)mosaic_destroy_datawidget, dialog);
 gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
                      GTK_SIGNAL_FUNC(gtk_widget_destroyed), g_dialog);

 button = gtk_button_new_with_label (_("Close"));
 GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
 gtk_signal_connect (GTK_OBJECT (button), "clicked",
                     (GtkSignalFunc) mosaic_destroy_datawidget, dialog);
 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), button,
                     TRUE, TRUE, 0);
 gtk_widget_grab_default (button);
 gtk_widget_show (button);

 vadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));
 vsb = gtk_vscrollbar_new (vadj);
 text = gtk_text_new (NULL, vadj);
 gtk_text_set_editable (GTK_TEXT (text), FALSE);
 gtk_widget_set_usize (text, 400, 475);

 table  = gtk_table_new (1, 2, FALSE);
 gtk_table_set_col_spacing (GTK_TABLE (table), 0, 2);

 gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), table, TRUE, TRUE, 0);

 gtk_table_attach (GTK_TABLE (table), vsb, 1, 2, 0, 1,
                   0, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
 gtk_table_attach (GTK_TABLE (table), text, 0, 1, 0, 1,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);

 gtk_container_set_border_width (GTK_CONTAINER (table), 2);

 font_strong =
   gdk_font_load (_("-*-helvetica-bold-r-normal-*-*-120-*-*-*-*-*-*"));
 font_emphasis =
   gdk_font_load (_("-*-helvetica-medium-o-normal-*-*-100-*-*-*-*-*-*"));
 font =
   gdk_font_load (_("-*-helvetica-medium-r-normal-*-*-100-*-*-*-*-*-*"));

 /*  Realize the widget before allowing new text to be inserted  */
 gtk_widget_realize (text);

 sprintf (buffer, "%s (v%s)\n\n",
                  _("Image Mosaic Plug-In for The GIMP"), VERSIO);
 gtk_text_insert (GTK_TEXT (text), font_strong, NULL, NULL, buffer, -1);
 gtk_text_insert (GTK_TEXT (text), font_emphasis, NULL, NULL,
                  "Copyright (C) 1999 Peter Kirchgessner\n\n", -1);
 gtk_text_insert (GTK_TEXT (text), font, NULL, NULL, gettext (helptext_1), -1);
 gtk_text_insert (GTK_TEXT (text), font, NULL, NULL, gettext (helptext_2), -1);

 gtk_widget_show (vsb);
 gtk_widget_show (text);
 gtk_widget_show (table);
}


static GtkWidget *
mosaic_make_fsel (GtkWidget **g_dialog,
                  char *title,
                  char *init_filename,
                  int hide_fileops,
                  void (*browse_ok_callback)(GtkWidget *, gpointer),
                  gpointer browsedata)

{GtkWidget *dialog;

 *g_dialog = dialog = gtk_file_selection_new (title);

 if (hide_fileops)
   gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (dialog));

 gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);

 gtk_file_selection_set_filename(GTK_FILE_SELECTION(dialog), init_filename);

 gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (dialog)->ok_button),
                     "clicked", (GtkSignalFunc) browse_ok_callback,
                     browsedata);
 gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (dialog)->cancel_button),
                     "clicked", (GtkSignalFunc)mosaic_destroy_datawidget,
                     dialog);
 gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
                      GTK_SIGNAL_FUNC(gtk_widget_destroyed), g_dialog);
 return dialog;
}


static void
mosaic_iscale_update (GtkAdjustment *adjustment,
                      int *scale_val)
{
 *scale_val = (int)adjustment->value;
}


static void
mosaic_browse_callback (GtkWidget *widget,
                        gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;
 static GtkWidget *dialog = NULL;

 if (!dialog)
 {
   mosaic_ui->mc_fsel =
     mosaic_make_fsel (&dialog, _("Select Mosaic Archive"),
                       mosaic_ui->create_vals->mia_name, 1,
                       mosaic_browse_ok_callback,
                       mosaic_ui);
 }
 if (!GTK_WIDGET_VISIBLE (dialog))
   gtk_widget_show (dialog);
 else
   gtk_widget_destroy (dialog);
}


static void
mosaic_browse_ok_callback (GtkWidget *widget,
                           gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;

 gtk_entry_set_text (GTK_ENTRY (mosaic_ui->mc_entry[3]),
   gtk_file_selection_get_filename (GTK_FILE_SELECTION (mosaic_ui->mc_fsel)));
 gtk_widget_destroy (GTK_WIDGET (mosaic_ui->mc_fsel));
}


static void
mia_create_callback (GtkWidget *widget,
                     gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;
 MiaCreateVals *miacreate_vals = mosaic_ui->miacreate_vals;
 GimpParam *return_vals;
 int nreturn_vals;

 /* Read mosaic archive name */
 strncpy (miacreate_vals->mia_name,
          gtk_entry_get_text (GTK_ENTRY (mosaic_ui->mia_entry[0])),
          sizeof (miacreate_vals->mia_name));
 miacreate_vals->mia_name[sizeof (miacreate_vals->mia_name)-1] = '\0';

 if (check_miacreate_vals (miacreate_vals) == 0)
 {
   return_vals = gimp_run_procedure ("plug_in_pmosaic_mia_create",
                    &nreturn_vals,
                    GIMP_PDB_INT32, GIMP_RUN_INTERACTIVE,
                    GIMP_PDB_STRING, miacreate_vals->mia_name,
                    GIMP_PDB_END);
   gimp_destroy_params (return_vals, nreturn_vals);
 }
}


static void
mia_browse_callback (GtkWidget *widget,
                     gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;
 static GtkWidget *dialog = NULL;

 if (!dialog)
 {
   mosaic_ui->mia_fsel =
     mosaic_make_fsel (&dialog, _("New Mosaic Archive"),
                       mosaic_ui->miacreate_vals->mia_name, 1,
                       mia_browse_ok_callback,
                       mosaic_ui);
 }
 if (!GTK_WIDGET_VISIBLE (dialog))
   gtk_widget_show (dialog);
 else
   gtk_widget_destroy (dialog);
}


static void
mia_browse_ok_callback (GtkWidget *widget,
                        gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;

 gtk_entry_set_text (GTK_ENTRY (mosaic_ui->mia_entry[0]),
   gtk_file_selection_get_filename (GTK_FILE_SELECTION (mosaic_ui->mia_fsel)));
 gtk_widget_destroy (GTK_WIDGET (mosaic_ui->mia_fsel));
}


static void
mini_create_callback (GtkWidget *widget,
                      gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;
 MiniCreateVals *minicreate_vals = mosaic_ui->minicreate_vals;
 GimpParam *return_vals;
 int nreturn_vals, sz;
 gint32 data_array[3];

 /* Read source directory */
 sz = sizeof (minicreate_vals->src_dir);
 strncpy (minicreate_vals->src_dir,
          gtk_entry_get_text (GTK_ENTRY (mosaic_ui->mini_entry[0])), sz);
 minicreate_vals->src_dir[sz-1] = '\0';

 /* Read destination directory */
 sz = sizeof (minicreate_vals->dst_dir);
 strncpy (minicreate_vals->dst_dir,
          gtk_entry_get_text (GTK_ENTRY (mosaic_ui->mini_entry[1])), sz);
 minicreate_vals->dst_dir[sz-1] = '\0';

 minicreate_vals->maxsize = mosaic_ui->mini_scale_value[0];
 minicreate_vals->quality = mosaic_ui->mini_scale_value[1];
 minicreate_vals->jpgppm_only = mosaic_ui->mini_jpgppm_only;

 data_array[0] = minicreate_vals->maxsize;
 data_array[1] = minicreate_vals->quality;
 data_array[2] = minicreate_vals->jpgppm_only;

 if (check_minicreate_vals (minicreate_vals) == 0)
 {
   return_vals = gimp_run_procedure ("plug_in_pmosaic_mini_create",
                    &nreturn_vals,
                    GIMP_PDB_INT32, GIMP_RUN_INTERACTIVE,
                    GIMP_PDB_STRING, minicreate_vals->src_dir,
                    GIMP_PDB_STRING, minicreate_vals->dst_dir,
                    GIMP_PDB_INT32, 3,
                    GIMP_PDB_INT32ARRAY, data_array,
                    GIMP_PDB_END);
   gimp_destroy_params (return_vals, nreturn_vals);
 }
}


static void
mini_browse_src_callback (GtkWidget *widget,
                          gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;
 static GtkWidget *dialog = NULL;

 if (!dialog)
 {
   mosaic_ui->mini_fsel[0] =
     mosaic_make_fsel (&dialog, _("Select Images Source Directory"),
                       mosaic_ui->minicreate_vals->src_dir, 1,
                       mini_browse_src_ok_callback,
                       mosaic_ui);
 }
 if (!GTK_WIDGET_VISIBLE (dialog))
   gtk_widget_show (dialog);
 else
   gtk_widget_destroy (dialog);
}


static void
mini_browse_src_ok_callback (GtkWidget *widget,
                             gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;

 gtk_entry_set_text (GTK_ENTRY (mosaic_ui->mini_entry[0]),
  gtk_file_selection_get_filename(GTK_FILE_SELECTION(mosaic_ui->mini_fsel[0])));
 gtk_widget_destroy (GTK_WIDGET (mosaic_ui->mini_fsel[0]));
}


static void
mini_browse_dst_callback (GtkWidget *widget,
                          gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;
 static GtkWidget *dialog = NULL;

 if (!dialog)
 {
   mosaic_ui->mini_fsel[1] =
     mosaic_make_fsel (&dialog, _("Select Thumbnails Destination Directory"),
                       mosaic_ui->minicreate_vals->dst_dir, 0,
                       mini_browse_dst_ok_callback,
                       mosaic_ui);
 }
 if (!GTK_WIDGET_VISIBLE (dialog))
   gtk_widget_show (dialog);
 else
   gtk_widget_destroy (dialog);
}


static void
mini_browse_dst_ok_callback (GtkWidget *widget,
                             gpointer data)

{MosaicInterface *mosaic_ui = (MosaicInterface *)data;

 gtk_entry_set_text (GTK_ENTRY (mosaic_ui->mini_entry[1]),
  gtk_file_selection_get_filename(GTK_FILE_SELECTION(mosaic_ui->mini_fsel[1])));
 gtk_widget_destroy (GTK_WIDGET (mosaic_ui->mini_fsel[1]));
}
