/*
 * Image mosaic software (C) 1999 Peter Kirchgessner
 * e-mail: peter@kirchgessner.net, WWW: http://www.kirchgessner.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define VERSIO "1.09"
/* Event history:
 * V 1.00, PK, 20-Feb-99, Creation
 * V 1.01, PK, 24-Mar-99, Save archive images as PPM for quality = 100.
 *                        Use better (but slower) scale algorithm.
 *                        Fix problem with endless loop when reading some
 *                        strange JPEG-files.
 * V 1.02, PK, 01-May-99, mirror images
 * V 1.03, PK, 31-Jul-99, add user defined file load
 *                        Dont use images with less than 5 pixel width/height
 * V 1.04, PK, 01-Jun-00, fix problem with recognition of archive signature
 *                        (could not load image archive)
 * V 1.05, PK, 06-Jun-00, Win32 port I
 * V 1.06, PK, 02-Oct-00, Win32 port II
 * V 1.07, PK, 27-Dec-00, Check for reusing images
 * V 1.08, PK, 23-Jan-01, Fix problem with reading non JPEG/PPM-files
 *                        on Win32
 * V 1.09, PK, 17-Nov-01, Add function to print list of used tiles
 */
static char ident[] = "@(#) ImageMosaic library v1.09 17-Nov-2001";

/* To use pmsc.c and pmsc.h as standalone program, define MSC_STAND_ALONE. */
/* Therefore these files should not use any GIMP-related definitions/includes */
/* #define MSC_STAND_ALONE */

#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef G_OS_WIN32
#include <unistd.h>
#include <dirent.h>
#else
#include <direct.h>
#include <io.h>
#endif

#include <jpeglib.h>

/* #define MSC_SCALE_QUICK */

#ifdef MSC_STAND_ALONE
#define MSC_ADD_MAIN
#endif

#include "pmsc.h"

#ifndef ABS
#define ABS(a) (((a) < 0) ? -(a) : (a))
#endif

#define MSC_MAX_PATH 1024

/* The maximum size for a mosaic file */
#define MOSAIC_MAX_SIZE   255

#define MAX_RGB_DIST      140

#define HUE_WEIGHT          2
#define MAX_HUE_DIST      255
#define MAX_SAT_DIST      255
#define MAX_VAL_DIST      255
#define MAX_HSV_DIST      600

/* Mosaic archive signature */
#define MSC_SIGNATURE "MSCA1;"

/* Number of tiles for which we have information about the mosaic file */
#define MOSAIC_TILES_HZ 5
#define MOSAIC_TILES_VT 5

/* The color information per file */
typedef struct
{
 unsigned char r, g, b, grey,   /* Total average RGB and grey-flag */
               tr[MOSAIC_TILES_HZ][MOSAIC_TILES_VT], /* Average red/tile */
               tg[MOSAIC_TILES_HZ][MOSAIC_TILES_VT], /* Average green/tile */
               tb[MOSAIC_TILES_HZ][MOSAIC_TILES_VT]; /* Average blue/tile */
} MOSAIC_COLOR_INFO;

/* All informations about a mosaic file */
typedef struct
{
 MOSAIC_IMAGE image;
 MOSAIC_COLOR_INFO cinfo;   /* color information */
 time_t mtime;              /* time of last modification */
 char *subdir;              /* subdirectory */
 char *name;                /* file name (base name only) */
} MOSAIC_FILE;

typedef enum
{
 MOSAIC_LANDSCAPE = 0, MOSAIC_PORTRAIT = 1, MOSAIC_SQUARE = 2
} MOSAIC_FORMAT;

typedef struct
{
  int avrg_width, avrg_height;
  int n;
  int sum_width, sum_height;
} MOSAIC_FORMAT_INFO;


/* Mosaic archive */
typedef struct mosaic_archive_s
{
 int size;
 int max_size;
 int increase;
 char *directory;
 int n_subdirs, max_subdirs;
 char **subdirs;
 MOSAIC_FORMAT_INFO format[3];
 MOSAIC_FILE *mf; /* Array of mosaic files without image data */
} MOSAIC_ARCHIVE;


/* Options for an archive image to replace a tile */
typedef struct {
  int index;
  int v_mirror;
} MOSAIC_BEST_FIT;

#define MOSAIC_IMAGE_USED_UNMIRRORED  1
#define MOSAIC_IMAGE_USED_V_MIRRORED  2

typedef struct msc_error_mgr {
  struct jpeg_error_mgr pub;    /* "public" fields */
  jmp_buf setjmp_buffer;        /* for return to caller */
} *msc_error_ptr;


typedef struct {
  int in_xpos, in_ypos;
  int in_width, in_height;
  int out_xpos, out_ypos;
  int out_width, out_height;
  MOSAIC_COLOR_INFO cinfo;
} TILE_INFO;


void (*msc_progress_function)(double) = NULL;
MOSAIC_MSGFUN msc_message_function = NULL;

/* Function declarations */
static void              msc_message (const char *msg);

static int               msc_rand (int k);
static char *            msc_strdup (const char *s);
static MOSAIC_IMAGE *    msc_image_alloc (int width, int height);
static void              msc_image_free (MOSAIC_IMAGE *img);

#ifdef MSC_NOT_USED
static MOSAIC_FILE *     msc_file_struct_alloc (void);
static void              msc_file_struct_free (MOSAIC_FILE *mf);
#endif
static void              msc_file_struct_comp_free (MOSAIC_FILE *mf);

static MOSAIC_ARCHIVE *  msc_archive_struct_alloc (void);
static void              msc_archive_struct_free (MOSAIC_ARCHIVE *ma);
static void              msc_archive_add_file_name (MOSAIC_ARCHIVE *ma,
                            MOSAIC_FILE *mf, const char *subdir,
                            const char *name);
static int               msc_archive_add_file_info (MOSAIC_ARCHIVE *ma,
                            const MOSAIC_FILE *mf);

static MOSAIC_FORMAT     msc_image_format (int width, int height);

static char            * msc_stripdirname (const char *dirname);
static int               msc_isdir (const char *name);
static int               msc_mkdir (const char *dirname);
static int               msc_mosaic_file_read (const char *dir,
                            const char *name, MOSAIC_FILE *mf);

static int               msc_jpg_image_read (const char *name,
                           MOSAIC_IMAGE *mi);
static int               msc_ppm_image_read (const char *name,
                           MOSAIC_IMAGE *mi);
static int               msc_jpg_image_write (const char *name,
                           const MOSAIC_IMAGE *mi, int quality);
static int               msc_ppm_image_write (const char *name,
                           const MOSAIC_IMAGE *mi);

static void              msc_color_info_v_mirror (
                           const MOSAIC_COLOR_INFO *cinfo,
                           MOSAIC_COLOR_INFO *cinfo_mirror);
static double            msc_rgb_color_distance (const MOSAIC_COLOR_INFO *ca,
                           const MOSAIC_COLOR_INFO *cb);
static void              msc_col_average_eval (const MOSAIC_IMAGE *image,
                            int xpos, int ypos, int width, int height,
                            unsigned char *r, unsigned char *g,
                            unsigned char *b);
static int               msc_color_is_grey (const MOSAIC_COLOR_INFO *cinfo);
static void              msc_color_info_eval (const MOSAIC_IMAGE *image,
                            MOSAIC_COLOR_INFO *cinfo);

static void              msc_dirname_get (const char *fullname, char *dir);
static void              msc_basename_get (const char *fullname, char *base);

static int               msc_find_scandir (void *userdata,
                             int (*action)(void *userdata,const char *subdir,
                                           const char *filename, int *abort),
                             const char *subdir);

static MOSAIC_ARCHIVE *  msc_archive_create (const char *basedir);
static int               msc_archive_write (const MOSAIC_ARCHIVE *ma,
                            const char *dirname, const char *name);
static MOSAIC_ARCHIVE *  msc_archive_read (const char *dirname,
                                           const char *name);

static void              msc_image_v_mirror (MOSAIC_IMAGE *img);
static int               msc_image_scale (MOSAIC_IMAGE *img_out,
                            const MOSAIC_IMAGE *img_in, int width, int height);
#ifdef MSC_STAND_ALONE
static void              msc_image_paste (MOSAIC_IMAGE *img_out,
                            const MOSAIC_IMAGE *img_in, int xpos, int ypos);
#endif
static void              msc_image_crop (MOSAIC_IMAGE *img_out,
                            const MOSAIC_IMAGE *img_in,
                            int xpos, int ypos, int width, int height);

static int               msc_archive_search_best_fit (const MOSAIC_ARCHIVE *mia,
                            const MOSAIC_COLOR_INFO *mia_hsv,
                            const MOSAIC_COLOR_INFO *cinfo,
                            const MOSAIC_COLOR_INFO *cinfo_hsv,
                            MOSAIC_FORMAT format, int exclude_grey,
                            int enable_v_mirror,
                            int reuse,
                            const char *used,
                            MOSAIC_BEST_FIT *best_fit);

static void              msc_mosaic_search_best_fit (const MOSAIC_ARCHIVE *mia,
                            int num_tiles, TILE_INFO *tile_info,
                            MOSAIC_FORMAT format, int exclude_grey,
                            int enable_v_mirror, int enable_random,
                            int reuse_image,
                            MOSAIC_BEST_FIT *best_fit);

static void               msc_mosaic_print_list (int n_rows, int n_columns,
                            int num_tiles,
                            const MOSAIC_BEST_FIT *best_fit,
                            const MOSAIC_ARCHIVE *mia);

/* Functions to handle image class */
#ifdef MSC_STAND_ALONE
static void       mscic_size_get (const MOSAIC_IMAGE_CLASS *this,
                              int *width, int *height);
static int        mscic_size_set (MOSAIC_IMAGE_CLASS *this,
                              int width, int height);
static void       mscic_crop (const MOSAIC_IMAGE_CLASS *this,
                              MOSAIC_IMAGE *img_out,
                              int xpos, int ypos, int width, int height);
static void       mscic_paste (MOSAIC_IMAGE_CLASS *this,
                              int xpos, int ypos,
                              const MOSAIC_IMAGE *img);
#endif
static int        mscic_paste_scaled (MOSAIC_IMAGE_CLASS *img_out,
                              int xpos, int ypos, int width, int height,
                              const MOSAIC_IMAGE *img_in);


/* Global error string */
static char msc_err_str[MSC_MAX_PATH];

static void
msc_message (const char *msg)

{
 if (msc_message_function)
   (*msc_message_function)(msg);
 else
   fprintf (stderr, "%s\n", msg);
}

/* Return pseudo-random value in range 0,..,k-1 */
static int
msc_rand (int k)

{
 return (int)(((double)k)*rand()/(RAND_MAX+1.0));
}

static char *
msc_strdup (const char *s)

{char *buf;

 if (s == NULL) return NULL;
 buf = malloc (strlen (s)+1);
 if (buf == NULL) return NULL;
 strcpy (buf, s);
 return buf;
}


static MOSAIC_IMAGE *
msc_image_alloc (int width, int height)

{MOSAIC_IMAGE *img = NULL;

 img = (MOSAIC_IMAGE *)calloc (1, sizeof (*img));
 if (img && (width > 0) && (height > 0))
 {
   img->width = width;
   img->height = height;
   img->rgb = (unsigned char *)malloc (width*height*3);
   if (img->rgb == NULL)
   {
     free (img);
     img = NULL;
   }
 }
 return img;
}


static void
msc_image_free (MOSAIC_IMAGE *img)

{
 if (img)
 {
   if (img->rgb)
   {
     free (img->rgb);
     img->rgb = NULL;
   }
   free (img);
 }
}


#ifdef MSC_NOT_USED
static MOSAIC_FILE *
msc_file_struct_alloc (void)

{
 return (MOSAIC_FILE *)calloc (1, sizeof (MOSAIC_FILE));
}


static void
msc_file_struct_free (MOSAIC_FILE *mf)

{
 if (mf)
 {
   msc_file_struct_comp_free (mf);
   free (mf);
 }
}
#endif


/* Free allocated components of a file struct */
static void
msc_file_struct_comp_free (MOSAIC_FILE *mf)

{
 if (mf)
 {
   if (mf->name)
   {
     free (mf->name);
     mf->name = NULL;
   }
   mf->subdir = NULL; /* Is allocated in archive */
   if (mf->image.rgb)
   {
     free (mf->image.rgb);
     mf->image.rgb = NULL;
   }
 }
}


static MOSAIC_ARCHIVE *
msc_archive_struct_alloc (void)

{MOSAIC_ARCHIVE *ma;

 ma = (MOSAIC_ARCHIVE *)calloc (1, sizeof (MOSAIC_ARCHIVE));
 if (ma)
 {
   ma->increase = 1024;
 }
 return ma;
}


static void
msc_archive_struct_free (MOSAIC_ARCHIVE *ma)

{int k;

 if (ma)
 {
   if (ma->directory)
   {
     free (ma->directory);
     ma->directory = NULL;
   }
   if (ma->mf)
   {
     for (k = 0; k < ma->size; k++)
       msc_file_struct_comp_free (&(ma->mf[k]));
     free (ma->mf);
     ma->mf = NULL;
   }
   if (ma->subdirs)
   {
     for (k = 0; k < ma->n_subdirs; k++)
     {
       free (ma->subdirs[k]);
       ma->subdirs[k] = NULL;
     }
     free (ma->subdirs);
     ma->subdirs = NULL;
   }
   free (ma);
 }
}


static void
msc_archive_add_file_name (MOSAIC_ARCHIVE *ma, MOSAIC_FILE *mf,
                           const char *subdir,
                           const char *name)

{char basename[MSC_MAX_PATH], dirname[MSC_MAX_PATH];
 char **dirlist;
 int k;

 mf->subdir = mf->name = NULL;

 /* If subdir specified, name keeps basename */
 if (subdir != NULL)
 {
   strcpy (dirname, subdir);
   strcpy (basename, name);
 }
 else /* name keeps full name */
 {
   msc_dirname_get (name, dirname);
   msc_basename_get (name, basename);
 }
 if ((dirname[0] == '.') && (dirname[1] == '/') && (dirname[2] != '\0'))
   memmove (dirname, dirname+2, strlen (dirname)-1);

 mf->name = msc_strdup (basename);
 if (mf->name == NULL) return;

 /* Search the subdirectory in the list */
 for (k = 0; k < ma->n_subdirs; k++)
 {
   if (strcmp (dirname, ma->subdirs[k]) == 0)
   {
     mf->subdir = ma->subdirs[k];
     return;
   }
 }

 /* We dont have it. Add it to the list of subdirectories */
 if (ma->n_subdirs >= ma->max_subdirs) /* Need more memory ? */
 {
   if (ma->max_subdirs == 0)
   {
     k = 64;
     ma->subdirs = (char **)malloc (k * sizeof (char *));
     if (ma->subdirs == NULL) goto cleanup;
     ma->max_subdirs = k;
   }
   else
   {
     k = ma->max_subdirs + 64;
     dirlist = (char **)realloc (ma->subdirs, k * sizeof (char *));
     if (dirlist == NULL) goto cleanup;
     ma->subdirs = dirlist;
     ma->max_subdirs = k;
   }
 }
 k = ma->n_subdirs;
 mf->subdir = ma->subdirs[k] = msc_strdup (dirname);
 if (ma->subdirs[k] != NULL)
 {
   (ma->n_subdirs)++;
   return;
 }

cleanup:
 if (mf->name)
 {
   free (mf->name);
   mf->name = NULL;
 }
 if (mf->subdir)
 {
   free (mf->subdir);
   mf->subdir = NULL;
 }
 return;
}


/* Add file information to an archive */
static int
msc_archive_add_file_info (MOSAIC_ARCHIVE *ma, const MOSAIC_FILE *mf)

{MOSAIC_FILE *new_mf;
 MOSAIC_FORMAT_INFO *format;
 int nbytes;

 if ((ma == NULL) || (mf == NULL)) return -1;
 if (ma->size+1 > ma->max_size)   /* Need additional memory ? */
 {
   nbytes = (ma->size + ma->increase) * sizeof (MOSAIC_FILE);
   if (ma->size == 0)
     new_mf = (MOSAIC_FILE *)malloc (nbytes);
   else
     new_mf = (MOSAIC_FILE *)realloc (ma->mf, nbytes);
   if (new_mf == NULL) return -1;
   ma->mf = new_mf;
   ma->max_size += ma->increase;
 }
 /* Copy the structure, add memory for name, drop image data */
 new_mf = &(ma->mf[ma->size]);
 memcpy (new_mf, mf, sizeof (*new_mf));

 msc_archive_add_file_name (ma, new_mf, mf->subdir, mf->name);
 if (new_mf->name == NULL) return -1;

 new_mf->image.rgb = NULL;

 /* Update information about formats kept in the archive */
 format = &(ma->format[msc_image_format (mf->image.width, mf->image.height)]);
 format->sum_width += mf->image.width;
 format->sum_height += mf->image.height;
 (format->n)++;
 format->avrg_width = format->sum_width / format->n;
 format->avrg_height = format->sum_height / format->n;

 (ma->size)++;

 return 0;
}


static void
msc_error_exit (j_common_ptr cinfo)

{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  msc_error_ptr mscerr = (msc_error_ptr) cinfo->err;

  /* Dont display the message. */
  /* (*cinfo->err->output_message) (cinfo); */

  /* Return control to the setjmp point */
  longjmp (mscerr->setjmp_buffer, 1);
}


static MOSAIC_FORMAT
msc_image_format (int width, int height)

{
 if (width*100 > height*110) return MOSAIC_LANDSCAPE;
 else if (height*100 > width*110) return MOSAIC_PORTRAIT;
 else return MOSAIC_SQUARE;
}


/* Make a copy of the string and strip off trailing directory separators */
static char *
msc_stripdirname (const char *dirname)

{char *ldir;
 int len;

 ldir = msc_strdup (dirname);
 if (ldir == NULL) return ldir;
 len = strlen (ldir);

#ifdef G_OS_WIN32
 /* We must not strip off for I:\ */
 if ((len > 3) && ((ldir[len-1] == '/') || (ldir[len-1] == '\\')))
   ldir[len-1] = '\0';
#else
 if (ldir[len-1] == '/')
   ldir[len-1] = '\0';
#endif

 return ldir;
}


/* Check if name is a directory */
static int
msc_isdir (const char *name)

{struct stat buf;

 if (name == NULL) return 0;
 if (stat (name, &buf) != 0) return 0;
#ifdef G_OS_WIN32
 return ((buf.st_mode & _S_IFDIR) ? 1 : 0);
#else
 return (S_ISDIR(buf.st_mode)) ? 1 : 0;
#endif
}


/* Create a directory (includeing top-directories). */
/* It is ok if the directory already exists */
static int
msc_mkdir (const char *dirname)

{char buffer[MSC_MAX_PATH];
 int is_dir, retval;

 /* Check if the parent directory is available */
 msc_dirname_get (dirname, buffer);
 is_dir =   (strcmp (buffer, ".") == 0)
          || (strcmp (buffer, "/") == 0)
          || msc_isdir (buffer);
 if (!is_dir)  /* If not available, create it */
 {
   retval = msc_mkdir (buffer);
   if (retval != 0) return retval;
 }
#ifndef G_OS_WIN32
 return mkdir (dirname, 0755);  /* Create the directory */
#else
 return mkdir (dirname);
#endif
}


/* Read a mosaic file */
static int
msc_mosaic_file_read (const char *dir, const char *name, MOSAIC_FILE *mf)

{int finished = 0, retval;
 char fullname[MSC_MAX_PATH];
 unsigned char *rgb = (unsigned char *)ident;
 struct stat statbuf;

 sprintf (fullname, "%s/%s", dir, name);

 /* Clear structure but keep possible image data pointer */
 rgb = mf->image.rgb;
 mf->image.rgb = NULL;
 msc_file_struct_comp_free (mf);
 memset (mf, 0, sizeof (*mf));
 mf->image.rgb = rgb;

 while (!finished)
 {
   retval = msc_jpg_image_read (fullname, &(mf->image));
   if (retval != 0) retval = msc_ppm_image_read (fullname, &(mf->image));
   if (retval != 0) break;

   mf->subdir = NULL;
   mf->name = msc_strdup (fullname);
   if (mf->name == NULL)
   {
     strcpy (msc_err_str, "msc_mosaic_file_read: Not enough memory");
     break;
   }
   if (stat (fullname, &statbuf) == 0)
     mf->mtime = statbuf.st_mtime;

   if ((mf->image.width < 2) || (mf->image.height < 2))
   {
     strcpy (msc_err_str, "msc_mosaic_file_read: Image too small");
     break;
   }
   if (   (mf->image.width > MOSAIC_MAX_SIZE)
       || (mf->image.height > MOSAIC_MAX_SIZE))
   {
     strcpy (msc_err_str, "msc_mosaic_file_read: Image too large");
     break;
   }
   finished = 1;
 }

 if (!finished)
 {
   if (rgb) mf->image.rgb = NULL;
   msc_file_struct_comp_free (mf);
   memset (mf, 0, sizeof (*mf));
   mf->image.rgb = rgb;
 }

 return finished ? 0 : -1;
}


/* Read a jpeg image. The way to read JPEG-files */
/* has been taken from the JPEG-plug-in */
static int volatile jpg_finished, jpg_init_cinfo, jpg_setjump;
static unsigned char *jpg_image;

static int
msc_jpg_image_read (const char *name, MOSAIC_IMAGE *mi)

{FILE *fd = fopen (name, "rb");
 struct jpeg_decompress_struct cinfo;
 struct msc_error_mgr jerr;
 int start_compress = 0;
 int keep_rgb = (mi->rgb != NULL);
 unsigned char *rowptr;

 jpg_image = NULL;
 jpg_finished = 0;
 jpg_init_cinfo = 0;
 jpg_setjump = 0;

 while (!jpg_finished)
 {
   if (fd == NULL)
   {
     sprintf (msc_err_str, "msc_jpg_image_read: Error opening file %s", name);
     break;
   }

   cinfo.err = jpeg_std_error (&jerr.pub);
   jerr.pub.error_exit = msc_error_exit;

   if (setjmp (jerr.setjmp_buffer))
   {
     jpg_setjump++;
     break;
   }

   /* initialize the JPEG decompression object */
   jpeg_create_decompress (&cinfo);
   jpg_init_cinfo = 1;

   /* specify data source (eg, a file) */
   jpeg_stdio_src (&cinfo, fd);

   /* read file parameters with jpeg_read_header() */
   (void) jpeg_read_header (&cinfo, TRUE);

   /* Start decompressor */
   jpeg_start_decompress (&cinfo);
   start_compress = 1;

   if ((cinfo.output_width < 1) || (cinfo.output_height < 1))
   {
     strcpy (msc_err_str, "msc_jpg_image_read: Image too small");
     break;
   }

   if (cinfo.output_components != 3)
   {
     sprintf (msc_err_str, "msc_jpg_image_read: Not a RGB file (%s)", name);
     break;
   }

   if (keep_rgb)
     jpg_image = mi->rgb;
   else
     jpg_image = malloc (3 * cinfo.output_width * cinfo.output_height);
   if (jpg_image == NULL)
   {
     strcpy (msc_err_str, "msc_jpg_image_read: Not enough memory");
     break;
   }

   /* Read scanlines */
   rowptr = jpg_image;
   while (cinfo.output_scanline < cinfo.output_height)
   {
     jpeg_read_scanlines (&cinfo, (JSAMPARRAY)&rowptr, 1);
     rowptr += cinfo.output_width*3;
   }

   mi->width = cinfo.output_width;
   mi->height = cinfo.output_height;
   mi->rgb = jpg_image;
   jpg_image = NULL;

   jpg_finished = 1;
 }
 if (jpg_finished && (jpg_setjump == 0)) jpeg_finish_decompress (&cinfo);
 if (jpg_init_cinfo) jpeg_destroy_decompress (&cinfo);
 if ((!keep_rgb) && (jpg_image != NULL)) free (jpg_image);
 if (fd) fclose (fd);

 return jpg_finished ? 0 : -1;
}


/* Read a ppm image */
static int
msc_ppm_image_read (const char *name, MOSAIC_IMAGE *mi)

{FILE *fd = fopen (name, "rb");
 char buf[MSC_MAX_PATH];
 int maxval, nbytes, nread, pgm, finished = 0;

 while (!finished)
 {
   if (fd == NULL)
   {
     sprintf (msc_err_str, "msc_ppm_image_read: Error opening file %s", name);
     break;
   }

   mi->width = mi->height = 0;

   if (fgets (buf, sizeof (buf), fd) == NULL)
   {
     strcpy (msc_err_str, "msc_ppm_image_read: Error reading header");
     break;
   }
   if ((buf[0] != 'P') || ((buf[1] != '5') && (buf[1] != '6')))
   {
     strcpy (msc_err_str, "msc_ppm_image_read: Not a PPM file");
     break;
   }
   pgm = (buf[1] == '5');
   do   /* Read width/height */
   {
     buf[0] = '\0';
     if (fgets (buf, sizeof (buf), fd) == NULL) break;
   } while (buf[0] == '#');
   if (sscanf (buf, "%d%d", &(mi->width), &(mi->height)) != 2)
   {
     strcpy (msc_err_str, "msc_ppm_image_read: Error reading width/height");
     break;
   }
   if ((mi->width < 1) || (mi->height < 1))
   {
     strcpy (msc_err_str, "msc_ppm_image_read: Invalid width/height");
     break;
   }

   do   /* Read maxval */
   {
     buf[0] = '\0';
     if (fgets (buf, sizeof (buf), fd) == NULL) break;
   } while (buf[0] == '#');
   if (sscanf (buf, "%d", &maxval) != 1)
   {
     strcpy (msc_err_str, "msc_ppm_image_read: Error reading maxval");
     break;
   }
   if (maxval != 255)
   {
     strcpy (msc_err_str, "msc_ppm_image_read: Bad maxval (not 255)");
     break;
   }
   nread = nbytes = mi->width * mi->height * 3;
   if (mi->rgb == NULL)
     mi->rgb = (unsigned char *)malloc (nbytes);
   if (mi->rgb == NULL)
   {
     strcpy (msc_err_str,"msc_ppm_image_read: Not enough memory to load image");
     break;
   }
   if (pgm) nread /= 3;
   if (fread (mi->rgb, 1, nread, fd) != nread)
   {
     strcpy (msc_err_str, "msc_ppm_image_read: Read error for image data");
     break;
   }
   if (pgm)   /* Expand to RGB ? */
   {register unsigned char *src, *dst;

     src = mi->rgb + nread;
     dst = mi->rgb + nbytes - 1;
     while (dst > src)
     {
       dst[0] = dst[-1] = dst[-2] = *(--src);
       dst -= 3;
     }
   }
   finished = 1;
 }
 if (fd) fclose (fd);

 return finished ? 0 : -1;
}


/* Write a jpeg image. The way to write JPEG-files */
/* has been taken from the JPEG-plug-in */
static int
msc_jpg_image_write (const char *name, const MOSAIC_IMAGE *mi, int quality)

{FILE * volatile outfile;
 struct jpeg_compress_struct cinfo;
 struct msc_error_mgr jerr;
 unsigned char *rowptr;

 jpg_image = NULL;
 jpg_finished = 0;
 jpg_init_cinfo = 0;

 cinfo.err = jpeg_std_error (&jerr.pub);
 jerr.pub.error_exit = msc_error_exit;

 outfile = NULL;

 /* Establish the setjmp return context for msc_error_exit to use. */
 if (setjmp (jerr.setjmp_buffer))
 {
   /* If we get here, the JPEG code has signaled an error.
    * We need to clean up the JPEG object, close the input file, and return.
    */
   jpeg_destroy_compress (&cinfo);
   if (outfile)
     fclose (outfile);

   return -1;
 }

 /* Initialize the JPEG compression object. */
 jpeg_create_compress (&cinfo);

 /* Specify data destination (eg, a file) */
 if ((outfile = fopen (name, "wb")) == NULL)
 {
   sprintf (msc_err_str, "msc_jpg_image_write: Error opening file %s", name);
   return -1;
 }
 jpeg_stdio_dest (&cinfo, outfile);

 cinfo.input_components = 3; /* number of color components per pixel */

 /* Set parameters for compression */
 cinfo.image_width = mi->width;
 cinfo.image_height = mi->height;

 /* Colorspace of input image */
 cinfo.in_color_space = JCS_RGB;

 /* Set default compression parameters */
 jpeg_set_defaults (&cinfo);

 /* Set own parameters */
 jpeg_set_quality (&cinfo, quality, 1);

 /* Start compressor */
 jpeg_start_compress (&cinfo, 1);

 while (cinfo.next_scanline < cinfo.image_height)
 {
   rowptr = mi->rgb + cinfo.next_scanline*mi->width*3;
   jpeg_write_scanlines (&cinfo, (JSAMPARRAY)&rowptr, 1);
 }

 /* Finish compression */
 jpeg_finish_compress (&cinfo);

 /* After finish_compress, we can close the output file. */
 fclose (outfile);

 /* Release JPEG compression object */
 /* This is an important step since it will release a good deal of memory. */
 jpeg_destroy_compress (&cinfo);

 return 0;
}


static int
msc_ppm_image_write (const char *name, const MOSAIC_IMAGE *mi)

{FILE *fp = fopen (name, "wb");
 int npix;

 if (fp == NULL)
 {
   msc_message ("msc_ppm_file_write: Error open file for writing");
   return -1;
 }
 fprintf (fp, "P6\n%d %d\n255\n", mi->width, mi->height);
 npix = mi->width * mi->height;
 if (fwrite (mi->rgb, 3, npix, fp) != npix)
 {
   msc_message ("msc_ppm_file_write: Write error");
   fclose (fp);
   return -1;
 }
 fclose (fp);
 return 0;
}


/* Create color info for a vertical mirrored image */
static void
msc_color_info_v_mirror (const MOSAIC_COLOR_INFO *cinfo,
                         MOSAIC_COLOR_INFO *cinfo_mirror)

{register int j, k;

 memcpy (cinfo_mirror, cinfo, sizeof (*cinfo_mirror));
 for (k = 0; k < MOSAIC_TILES_HZ; k++)
 {
   for (j = 0; j < MOSAIC_TILES_VT; j++)
   {
     cinfo_mirror->tr[k][j] = cinfo->tr[MOSAIC_TILES_HZ-k][j];
     cinfo_mirror->tg[k][j] = cinfo->tg[MOSAIC_TILES_HZ-k][j];
     cinfo_mirror->tb[k][j] = cinfo->tb[MOSAIC_TILES_HZ-k][j];
   }
 }
}


/* Calculate distance for two rgb color descriptions */
static double
msc_rgb_color_distance (const MOSAIC_COLOR_INFO *ca,
                        const MOSAIC_COLOR_INFO *cb)

{double sumrgb = 0.0;
 int dr, dg, db;
 int k, j;

 for (k = 0; k < MOSAIC_TILES_HZ; k++)
 {
   for (j = 0; j < MOSAIC_TILES_VT; j++)
   {
     dr = ((int)ca->tr[k][j]) - ((int)cb->tr[k][j]);
     dg = ((int)ca->tg[k][j]) - ((int)cb->tg[k][j]);
     db = ((int)ca->tb[k][j]) - ((int)cb->tb[k][j]);
     sumrgb += sqrt ((double)(dr*dr + dg*dg + db*db));
   }
 }
 return sumrgb / (double)(255.0*MOSAIC_TILES_HZ * MOSAIC_TILES_VT);
}


/* Evaluate average color info for an image. Only the rectangle
 * specified by xpos,ypos width,height is used
 */
static void
msc_col_average_eval (const MOSAIC_IMAGE *image,
                      int xpos, int ypos, int width, int height,
                      unsigned char *r, unsigned char *g, unsigned char *b)

{register const unsigned char *rowptr;
 register int rr, rg, rb;
 int j, k;

 *r = *g = *b = 0;
 if ((width <= 0) || (height <= 0)) return;

 rr = rg = rb = 0;
 for (j = 0; j < height; j++)
 {
   rowptr = image->rgb + ((ypos+j)*image->width*3) + xpos*3;
   for (k = 0; k < width; k++)
   {
     rr += *(rowptr++);
     rg += *(rowptr++);
     rb += *(rowptr++);
   }
 }
 k = width * height;
 *r = (unsigned char)(rr / k);
 *g = (unsigned char)(rg / k);
 *b = (unsigned char)(rb / k);
}


static int
msc_color_is_grey (const MOSAIC_COLOR_INFO *cinfo)

{int j, k;

 if ((ABS (cinfo->r - cinfo->g) > 7) || (ABS (cinfo->g - cinfo->b) > 7))
   return 0;

 for (j = 0; j < MOSAIC_TILES_HZ; j++)
   for (k = 0; k < MOSAIC_TILES_VT; k++)
    if (   (ABS (cinfo->tr[j][k] - cinfo->tg[j][k]) > 7)
        || (ABS (cinfo->tg[j][k] - cinfo->tb[j][k]) > 7))
      return 0;
 return 1;
}


/* Evaluate the color information for a mosaic file. */
static void
msc_color_info_eval (const MOSAIC_IMAGE *image,
                     MOSAIC_COLOR_INFO *cinfo)

{int j, k;
 int xpos, ypos, width, height;

 /* Get overall color average */
 msc_col_average_eval (image, 0, 0, image->width, image->height,
                       &(cinfo->r), &(cinfo->g), &(cinfo->b));

 width = image->width / MOSAIC_TILES_HZ;
 if (width < 1) width = 1;
 height = image->height / MOSAIC_TILES_VT;
 if (height < 1) height = 1;
 for (j = 0; j < MOSAIC_TILES_VT; j++)
 {
   ypos = (image->height * j) / MOSAIC_TILES_VT;
   for (k = 0; k < MOSAIC_TILES_HZ; k++)
   {
     xpos = (image->width * k) / MOSAIC_TILES_HZ;
     msc_col_average_eval (image, xpos, ypos, width, height,
                           &(cinfo->tr[j][k]), &(cinfo->tg[j][k]),
                           &(cinfo->tb[j][k]));
   }
 }
 cinfo->grey = (unsigned char)msc_color_is_grey (cinfo);
}


static void
msc_dirname_get (const char *fullname, char *dir)

{int lastidx = strlen (fullname) - 1;
 int firstidx = 0;

 /* find where the last directory separator is */
 while (lastidx >= 0)
 {
#ifdef G_OS_WIN32
   if ((fullname[lastidx] == '/') || (fullname[lastidx] == '\\')) break;
#else
   if (fullname[lastidx] == '/') break;
#endif
   lastidx--;
 }

 /* find where the first directory separator is */
 while (fullname [firstidx] != '\0')
 {
#ifdef G_OS_WIN32
   if ((fullname[firstidx] == '/') || (fullname[firstidx] == '\\')) break;
#else
   if (fullname[firstidx] == '/') break;
#endif
   firstidx++;
 }

 if (lastidx < 0)   /* No directory specified ? use current directory */
 {
   strcpy (dir, ".");
   return;
 }

 /* Root directory ? */
 if (lastidx == 0)
 {
   strcpy (dir, "/");
   return;
 }

 if (firstidx == lastidx)
 {
#ifdef G_OS_WIN32
   /* Could be c:\. Must keep backslash */
   strncpy (dir, fullname, lastidx+1);
   dir[lastidx+1] = '\0';
#else
   /* Could be localdir/filename. Strip off slash */
   strncpy (dir, fullname, lastidx);
   dir[lastidx] = '\0';
#endif
 }
 else   /* Some other directory. Strip off the slash */
 {
   strncpy (dir, fullname, lastidx);
   dir[lastidx] = '\0';
 }
}


static void
msc_basename_get (const char *fullname, char *base)

{int idx = strlen (fullname) - 1;

 while (idx >= 0)
 {
#ifdef G_OS_WIN32
   if (fullname[idx] == '\\') break;
#endif
   if (fullname[idx] == '/') break;
   idx--;
 }
 strcpy (base, fullname+idx+1);
}


/* Scan a directory for all files. Start at current directory */
static int
msc_find_scandir (void *userdata,
                  int (*action)(void *userdata, const char *subdir,
                                const char *filename, int *abort),
                  const char *subdir)

{
#ifndef G_OS_WIN32

 DIR *dir = NULL;
 struct dirent *dirent;
 char fullname[MSC_MAX_PATH];
 struct stat statbuf;
 int abort = 0;

 dir = opendir (subdir);
 if (dir == NULL) goto cleanup;

 while (!abort) /* For all entries in the directory */
 {
   dirent = readdir (dir);
   if (dirent == NULL) break;
   if (   (strcmp (dirent->d_name, ".") == 0)
       || (strcmp (dirent->d_name, "..") == 0)) continue;

   /* Might be a directory */
   sprintf (fullname, "%s/%s", subdir, dirent->d_name);
   if ((stat (fullname, &statbuf) == 0) && S_ISDIR(statbuf.st_mode))
   {
     if (msc_find_scandir (userdata, action, fullname) < 0)
       abort = 1;
     continue;
   }

   abort = 0;
   (*action)(userdata, subdir, dirent->d_name, &abort);
 }

cleanup:
 if (dir) closedir (dir);

 return abort ? -1 : 0;

#else

 /* This is for Windows */
 struct _finddata_t c_file;
 long fhand = -1;
 char fullname[MSC_MAX_PATH];
 char filepattern[MSC_MAX_PATH];
 int abort = 0;

 strcpy (filepattern, subdir);
 strcat (filepattern, "/*");

 fhand = _findfirst (filepattern, &c_file);
 if (fhand == -1) goto cleanup;

 while (!abort) /* For all entries in the directory */
 {
   /* Might be a directory */
   sprintf (fullname, "%s/%s", subdir, c_file.name);
   if (c_file.attrib & _A_SUBDIR)
   {
     if (   (strcmp (c_file.name, ".") != 0)
         && (strcmp (c_file.name, "..") != 0))
     {
       if (msc_find_scandir (userdata, action, fullname) < 0)
       {
         abort = 1;
         break;
       }
     }
   }
   else
   {
     abort = 0;
     (*action)(userdata, subdir, c_file.name, &abort);
     if (abort) break;
   }
   if (_findnext (fhand, &c_file) != 0) break;
 }

cleanup:
 if (fhand != -1) _findclose (fhand);

 return abort ? -1 : 0;

#endif
}


static int
msc_files_count_callback (void *userdata, const char *subdir,
                          const char *filename, int *abort)
{int *count = (int *)userdata;

 *abort = 0; /* Dont abort */

 (*count)++;
 return 0;
}


/* Count all files below a directory */
int
msc_files_count (const char *basedir)

{int count = 0;
 char *l_basedir;

 l_basedir = msc_stripdirname (basedir);

 if (msc_files_find (l_basedir, &count, msc_files_count_callback) != 0)
   count = -1;

 free (l_basedir);
 return count;
}


/* Do some action on all files of a directory incl. subdirectories. */
/* Returns 0 on success, -1 on failure */
int
msc_files_find (const char *basedir, void *userdata,
                int (*action)(void *userdata, const char *subdir,
                              const char *filename, int *abort))

{char buf[MSC_MAX_PATH], wd[MSC_MAX_PATH];
 int retval = -1;
 char *l_basedir;

 wd[0] = '\0';
 l_basedir = msc_stripdirname (basedir);

 /* Get working directory */
 if (getcwd (wd, sizeof (wd)) == NULL)
 {
   msc_message ("msc_files_find: Error getting working directory");
   goto cleanup;
 }
 if (chdir (l_basedir) != 0)
 {
   sprintf (buf, "msc_files_find: Cannot change to directory %s", l_basedir);
   msc_message (buf);
   goto cleanup;
 }

 retval = msc_find_scandir (userdata, action, ".");

cleanup:
 if (wd[0] != '\0') chdir (wd);

 free (l_basedir);
 return retval;
}


typedef struct {
 MOSAIC_ARCHIVE *ma;

 int maxfiles;
 int nfiles;
} MOSAIC_ADD_INFO;

/* Add file information to archive */
static int
msc_archive_create_callback (void *userdata,
                             const char *subdir,
                             const char *filename,
                             int *abort)

{MOSAIC_ADD_INFO *inf = (MOSAIC_ADD_INFO *)userdata;
 MOSAIC_ARCHIVE *ma = inf->ma;
 MOSAIC_FILE mf;
 int ev;

 *abort = 0;
 if (ma == NULL) { *abort = 1; return -1; }
 memset (&mf, 0, sizeof (mf));

 (inf->nfiles)++;

 /* Want a progress indicator ? */
 if (   (inf->maxfiles > 0) && (msc_progress_function != NULL)
     && ((inf->nfiles % 10) == 1))
 {double progress;

   progress = ((double)inf->nfiles) / (double)inf->maxfiles;
   if (progress > 1.0) progress = 1.0;
   (*msc_progress_function) (progress);
 }

 ev = msc_mosaic_file_read (subdir, filename, &mf);
 if (ev == 0)                       /* If the file is ok */
 {                                  /* get the color info */
   msc_color_info_eval (&(mf.image), &(mf.cinfo));
   msc_archive_add_file_info (ma, &mf);  /* Add info to archive */
   msc_file_struct_comp_free (&mf);      /* Free additional memory */
 }
 return 0;
}


/* Try to read all mosaic files of a directory */
MOSAIC_ARCHIVE *
msc_archive_create (const char *basedir)

{MOSAIC_ADD_INFO inf;
 MOSAIC_ARCHIVE *ma = NULL;
 int err = 1;

 ma = msc_archive_struct_alloc ();
 if (ma == NULL) goto cleanup;

 inf.ma = ma;
 if (msc_progress_function)
   inf.maxfiles = msc_files_count (basedir);
 else
   inf.maxfiles = 0;

 inf.nfiles = 0;

 if (msc_files_find (basedir, (void *)&inf, msc_archive_create_callback) != 0)
   goto cleanup;

 if (inf.maxfiles > 0)
   (*msc_progress_function) (1.0);

 if (ma->size > 0)   /* No files found ? Treat it as error */
   err = 0;

cleanup:
 if (err && (ma != NULL))
 {
   msc_archive_struct_free (ma);
   ma = NULL;
 }

 return ma;
}


static int
msc_archive_write (const MOSAIC_ARCHIVE *ma, const char *dirname,
                   const char *name)

{FILE *fd = NULL;
 char archive[MSC_MAX_PATH];
 int k, ind, err = 1;
 MOSAIC_FILE *mf;

#define MSC_WRITE_I8(ival) \
  { putc((ival) & 0xff, fd); }

#define MSC_WRITE_I16(ival) \
 {int iv = (int)(ival); putc((iv >> 8) & 0xff, fd); \
  putc(iv & 0xff, fd); }

#define MSC_WRITE_I32(ival) \
 {int iv = (int)(ival); putc((iv >> 24) & 0xff, fd); \
  putc((iv >> 16) & 0xff, fd); putc ((iv >> 8) & 0xff, fd); \
  putc(iv &0xff, fd); }

#define MSC_WRITE_STR(str) \
 { MSC_WRITE_I16(strlen (str)); \
   if (fwrite (str, 1, strlen (str), fd) != strlen (str)) goto cleanup; }

#define MSC_WRITE_MEM(ptr,size) \
 if (fwrite (ptr, 1, size, fd) != (size)) goto cleanup;


 sprintf (archive, "%s/%s", dirname, name);
 fd = fopen (archive, "wb");
 if (fd == NULL) goto cleanup;

 MSC_WRITE_MEM (MSC_SIGNATURE, strlen (MSC_SIGNATURE));
 MSC_WRITE_I32 (ma->size);

 for (k = 0; k < 3; k++)
 {
   MSC_WRITE_I32 (ma->format[k].avrg_width);
   MSC_WRITE_I32 (ma->format[k].avrg_height);
   MSC_WRITE_I32 (ma->format[k].n);
   MSC_WRITE_I32 (ma->format[k].sum_width);
   MSC_WRITE_I32 (ma->format[k].sum_height);
 }

 MSC_WRITE_I32 (ma->n_subdirs);
 for (k = 0; k < ma->n_subdirs; k++)
 {
   MSC_WRITE_STR (ma->subdirs[k]);
 }

 mf = &(ma->mf[0]);
 for (k = 0; k < ma->size; k++)
 {
   MSC_WRITE_I16 (mf->image.width);
   MSC_WRITE_I16 (mf->image.height);
   MSC_WRITE_I32 (mf->mtime);
   MSC_WRITE_MEM (&(mf->cinfo), sizeof (mf->cinfo));
   for (ind = 0; ind < ma->n_subdirs; ind++)
     if (strcmp (mf->subdir, ma->subdirs[ind]) == 0) break;
   MSC_WRITE_I32 (ind);
   MSC_WRITE_STR (mf->name);
   mf++;
 }
 err = 0;

cleanup:

 if (fd) fclose (fd);
 if (err) unlink (archive);
 return err ? -1 : 0;
}


int
msc_mia_create (const char *mia_name)

{char *dirname, *basename;
 int retval;
 MOSAIC_ARCHIVE *mia;

 if ((mia_name == NULL) || (mia_name[0] == '\0')) return -1;

 dirname = msc_strdup (mia_name);
 basename = msc_strdup (mia_name);
 if ((dirname == NULL) || (basename == NULL)) return -1;

 msc_dirname_get (mia_name, dirname);
 msc_basename_get (mia_name, basename);

 mia = msc_archive_create (dirname);
 if (mia == NULL) return -1;

 retval = msc_archive_write (mia, dirname, basename);

 msc_archive_struct_free (mia);
 free (basename);
 free (dirname);

 return retval;
}

typedef struct {
 const char *srcdir;
 const char *dstdir;
 int maxsize;
 int quality;
 int (*readfun) (const char *filename, MOSAIC_IMAGE *mi);

 int maxfiles;
 int nfiles;
} MOSAIC_COLLECT_INFO;


/* Add file information to archive */
static int
msc_images_collect_callback (void *userdata,
                             const char *subdir,
                             const char *filename,
                             int *abort)

{MOSAIC_COLLECT_INFO *inf = (MOSAIC_COLLECT_INFO *)userdata;
 MOSAIC_IMAGE mi, *dst_mi;
 char fullname[MSC_MAX_PATH], dirname[MSC_MAX_PATH];
 char abspath[MSC_MAX_PATH];
 int ev = 0, dst_width, dst_height;

 *abort = 0;
 memset (&mi, 0, sizeof (mi));
 if (strncmp (subdir, "./", 2) == 0) subdir += 2;

 (inf->nfiles)++;

 /* Want a progress indicator ? */
 if (   (inf->maxfiles > 0) && (msc_progress_function != NULL)
     && ((inf->nfiles % 10) == 1))
 {double progress;

   progress = ((double)inf->nfiles) / (double)inf->maxfiles;
   if (progress > 1.0) progress = 1.0;
   (*msc_progress_function) (progress);
 }

 sprintf (fullname, "%s/%s", subdir, filename);
 ev = msc_jpg_image_read (fullname, &mi);
 if (ev != 0)
 {
   ev = msc_ppm_image_read (fullname, &mi);
   if ((ev != 0) && (inf->readfun != NULL))
   {            /* Call user read function with absolute path to file */
     char *dirsep = "/", slash = 0;

     /* Check if srcdir already has the directory separator */
     if (strlen (inf->srcdir) > 0) slash = inf->srcdir[strlen (inf->srcdir)-1];
     if ((slash == '/') || (slash == '\\')) dirsep = "";
     if (strcmp (subdir, ".") == 0)
       sprintf (abspath, "%s%s%s", inf->srcdir, dirsep, filename);
     else
       sprintf (abspath, "%s%s%s/%s", inf->srcdir, dirsep, subdir, filename);

     ev = (inf->readfun) (abspath, &mi);
   }
 }
 if (ev == 0)                       /* If the file is ok */
 {MOSAIC_IMAGE scaled_mi;
  unsigned char rgb[MOSAIC_MAX_SIZE*MOSAIC_MAX_SIZE*3];

   if ((mi.width < MOSAIC_TILES_HZ) || (mi.height < MOSAIC_TILES_VT))
   {
     free (mi.rgb);
     return 1;   /* Treat it as read error */
   }

   /* Scale it if necessary*/
   if ((mi.width > inf->maxsize) || (mi.height > inf->maxsize))
   {
     if (mi.width >= mi.height)  /* Landscape format ? */
     {
       dst_width = inf->maxsize;
       dst_height = (inf->maxsize * mi.height) / mi.width;
     }
     else
     {
       dst_height = inf->maxsize;
       dst_width = (inf->maxsize * mi.width) / mi.height;
     }
     if ((dst_width < MOSAIC_TILES_HZ) || (dst_height < MOSAIC_TILES_VT))
     {
       free (mi.rgb);
       return 1;   /* Treat it as read error */
     }

     scaled_mi.width = scaled_mi.height = 0;
     scaled_mi.rgb = rgb;

     msc_image_scale (&scaled_mi, &mi, dst_width, dst_height);

     dst_mi = &scaled_mi;
   }
   else
   {
     dst_mi = &mi;
   }

   /* Write it below destination directory */
   /* Maybe we need to make the path */
   sprintf (dirname, "%s/%s", inf->dstdir, subdir);
   ev = 0;
   if (!msc_isdir (dirname))
   {
     ev = msc_mkdir (dirname);
     if (ev != 0)
     {
       sprintf (fullname, "Error creating directory %s", dirname);
       msc_message (fullname);
     }
   }
   if (ev == 0)
   {
     sprintf (fullname, "%s/%s/%s", inf->dstdir, subdir, filename);
     if (inf->quality >= 100) /* No loss in quality */
       ev = msc_ppm_image_write (fullname, dst_mi);
     else
       ev = msc_jpg_image_write (fullname, dst_mi, inf->quality);
     if (ev != 0)
     {
       sprintf (fullname, "Error writing mosaic archive image %s/%s/%s",
                inf->dstdir, subdir, filename);
       msc_message (fullname);
     }
   }
   free (mi.rgb);
   if (ev != 0) *abort = 1;  /* Abort on write errors */
 }
 /* Errors on reading are ignored */
 return ev;
}


/* Collect a number of images. Read all files below srcdir, */
/* Scale them down and write them below dstdir */
int
msc_images_collect (const char *srcdir,
                    const char *dstdir,
                    int maxsize,
                    int quality,
                    int (*user_read) (const char *filename, MOSAIC_IMAGE *mi))

{MOSAIC_COLLECT_INFO inf;
 char buffer[MSC_MAX_PATH];
 char *l_srcdir, *l_dstdir;

 l_srcdir = msc_stripdirname (srcdir);
 l_dstdir = msc_stripdirname (dstdir);

 if ((l_srcdir == NULL) || (l_dstdir == NULL)) return -1;

 if (!msc_isdir (l_srcdir))
 {
   sprintf (buffer, "msc_images_collect: Source directory %s not found",l_srcdir);
   msc_message (buffer);
   free (l_srcdir);
   free (l_dstdir);
   return -1;
 }
 if (!msc_isdir (l_dstdir))
 {
   sprintf (buffer, "msc_images_collect: Destination directory %s not found",
            l_dstdir);
   msc_message (buffer);
   free (l_srcdir);
   free (l_dstdir);
   return -1;
 }
 if (maxsize < 32) maxsize = 32;
 else if (maxsize > MOSAIC_MAX_SIZE) maxsize = MOSAIC_MAX_SIZE;
 if (quality < 10) quality = 10;
 else if (quality > 100) quality = 100;
 inf.srcdir = l_srcdir;
 inf.dstdir = l_dstdir;
 inf.maxsize = maxsize;
 inf.quality = quality;
 inf.readfun = user_read;

 if (msc_progress_function)
   inf.maxfiles = msc_files_count (l_srcdir);
 else
   inf.maxfiles = 0;

 inf.nfiles = 0;

 if (msc_files_find (l_srcdir, (void *)&inf, msc_images_collect_callback) != 0)
 {
  inf.nfiles = -1;
 }
 else if (inf.maxfiles > 0)
   (*msc_progress_function) (1.0);

 free (l_srcdir);
 free (l_dstdir);

 return inf.nfiles;
}

static MOSAIC_ARCHIVE *
msc_archive_read (const char *dirname, const char *name)

{FILE *fd = NULL;
 char archive[MSC_MAX_PATH], buf[MSC_MAX_PATH];
 int k, ind, err = 1;
 unsigned long ul;
 MOSAIC_FILE *mf;
 MOSAIC_ARCHIVE *ma = NULL;

 sprintf (archive, "%s/%s", dirname, name);

 fd = fopen (archive, "rb");
 if (fd == NULL)
 {
   sprintf (msc_err_str, "msc_archive_read: Error opening %s", archive);
   goto cleanup;
 }

 ma = msc_archive_struct_alloc ();
 if (ma == NULL)
 {
   sprintf (msc_err_str, "msc_archive_read: Not enough memory");
   goto cleanup;
 }

#define MSC_READ_MEM(ptr,size) \
 if (fread (ptr, 1, size, fd) != (size)) \
 { strcpy (msc_err_str, "msc_archive_read: Read error"); goto cleanup; }

#define MSC_READ_I8(ptr,s) \
 {int iv; iv = getc (fd); \
  if (iv == EOF) \
  { sprintf (msc_err_str, "msc_archive_read: Error reading i8 (%s)", s);\
   goto cleanup; } \
  *(ptr) = iv; }

#define MSC_READ_I16(ptr,s) \
 {int iv1, iv2; iv1 = getc (fd); iv2 = getc (fd); \
  if (iv2 == EOF) \
  { sprintf (msc_err_str, "msc_archive_read: Error reading i16 (%s)", s);\
   goto cleanup; } \
  *(ptr) = (iv1 << 8) + iv2; }

#define MSC_READ_I32(ptr,s) \
 {int iv1, iv2, iv3, iv4; iv1 = getc (fd); iv2 = getc (fd); \
  iv3 = getc (fd); iv4 = getc (fd); \
  if (iv4 == EOF) \
  { sprintf (msc_err_str, "msc_archive_read: Error reading i32 (%s)", s);\
   goto cleanup; } \
  *(ptr) = (iv1 << 24) + (iv2 << 16) + (iv3 << 8) + iv4; }

#define MSC_READ_STR(ptr,s) \
 {int sl; MSC_READ_I16(&sl,s); \
  if (sl >= sizeof(ptr)) \
  { sprintf (msc_err_str, "msc_archive_read: string too long (%s)", s);\
   goto cleanup; } \
  if (fread (ptr, 1, sl, fd) != sl) \
  { sprintf (msc_err_str, "msc_archive_read: Error reading string (%s)", s);\
   goto cleanup; } \
   (ptr)[sl] = '\0'; }


 MSC_READ_MEM (buf, strlen (MSC_SIGNATURE));
 /* Fix problem with recognition of archive. Use strncmp() */
 if (strncmp (buf, MSC_SIGNATURE, strlen (MSC_SIGNATURE)) != 0)
 {
   strcpy (msc_err_str, "msc_archive_read: Invalid signature");
   goto cleanup;
 }

 MSC_READ_I32 (&(ma->size), "size");

 ma->mf = (MOSAIC_FILE *)calloc (ma->size, sizeof (MOSAIC_FILE));
 if (ma->mf == NULL)
 {
   sprintf (msc_err_str, "msc_archive_read: Not enough memory");
   goto cleanup;
 }

 for (k = 0; k < 3; k++)
 {
   MSC_READ_I32 (&(ma->format[k].avrg_width), "average width");
   MSC_READ_I32 (&(ma->format[k].avrg_height), "average height");
   MSC_READ_I32 (&(ma->format[k].n), "format n");
   MSC_READ_I32 (&(ma->format[k].sum_width), "sum width");
   MSC_READ_I32 (&(ma->format[k].sum_height), "sum height");
 }

 MSC_READ_I32 (&(ma->max_subdirs), "number of subdirectories");
 ma->subdirs = (char **)calloc (ma->max_subdirs, sizeof (char *));
 if (ma->subdirs == NULL)
 {
   sprintf (msc_err_str, "msc_archive_read: Not enough memory (subdirs)");
   goto cleanup;
 }

 for (k = 0; k < ma->max_subdirs; k++)
 {
   MSC_READ_STR (buf, "subdir");
   ma->subdirs[k] = msc_strdup (buf);
   if (ma->subdirs[k] == NULL)
   {
     sprintf (msc_err_str, "msc_archive_read: Not enough memory (subdir)");
     goto cleanup;
   }
 }
 ma->n_subdirs = ma->max_subdirs;

 mf = &(ma->mf[0]);
 for (k = 0; k < ma->size; k++)
 {
   MSC_READ_I16 (&(mf->image.width), "width");
   MSC_READ_I16 (&(mf->image.height), "height");
   MSC_READ_I32 (&ul, "mtime");
   mf->mtime = (time_t)ul;
   MSC_READ_MEM (&(mf->cinfo), sizeof (mf->cinfo));
   MSC_READ_I32 (&ind, "subdir index");
   mf->subdir = ma->subdirs[ind];
   MSC_READ_STR (buf, "subdir");
   mf->name = msc_strdup (buf);
   if (mf->name == NULL)
   {
     sprintf (msc_err_str, "msc_archive_read: Not enough memory");
     goto cleanup;
   }
   mf++;
 }
 err = 0;
 ma->directory = msc_strdup (dirname);

cleanup:

 if (fd) fclose (fd);
 if (err && (ma != NULL))
 {
   msc_archive_struct_free (ma);
   ma = NULL;
 }
 return ma;
}


/* Mirror image vertically */
static void
msc_image_v_mirror (MOSAIC_IMAGE *img)

{register unsigned char *rgb_in, *rgb_out, tmp;
 register int x, y;

 for (y = 0; y < img->height; y++)
 {
   rgb_in = img->rgb + y*img->width*3;
   rgb_out = rgb_in + (img->width-1)*3;
   x = img->width/2;
   while (x-- > 0)
   {
     tmp = *rgb_in; *(rgb_in++) = *rgb_out; *(rgb_out++) = tmp;
     tmp = *rgb_in; *(rgb_in++) = *rgb_out; *(rgb_out++) = tmp;
     tmp = *rgb_in; *(rgb_in++) = *rgb_out; *(rgb_out++) = tmp;
     rgb_out -= 6;
   }
 }
}


#ifdef MSC_SCALE_QUICK

/* Quick and dirty scale routine. The memory */
/* for the output image must already be allocated */
static int
msc_image_scale (MOSAIC_IMAGE *img_out, const MOSAIC_IMAGE *img_in,
                int width, int height)

{register int *rowoffset;
 register unsigned char *rgb_in, *rgb_out;
 register unsigned char *r;
 int dx, dy;

 dx = img_in->width - width;
 dy = img_in->height - height;

 /* Output image a little bit smaller ? Crop it */
 if ((dx >= 0) && (dx <= 5) && (dy >= 0) && (dy <= 5))
 {
   msc_image_crop (img_out, img_in, dx/2, dy/2, width, height);
   return 0;
 }

 rowoffset = malloc (width * sizeof (int));
 if (rowoffset == NULL) return -1;

 for (dx = 0; dx < width; dx++)
   rowoffset[dx] = ((dx * img_in->width) / width)*3;

 img_out->width = width;
 img_out->height = height;
 for (dy = 0; dy < height; dy++)
 {
   rgb_in = img_in->rgb + ((dy * img_in->height)/height)*3*img_in->width;
   rgb_out = img_out->rgb + dy * width * 3;

   for (dx = 0; dx < width; dx++)
   {
     r = rgb_in + rowoffset[dx];
     *(rgb_out++) = *(r++);
     *(rgb_out++) = *(r++);
     *(rgb_out++) = *r;
   }
 }
 free (rowoffset);
 return 0;
}

#else

static void
msc_row_scale_up (const unsigned char *in_urgb, int *out_irgb,
                  int out_width, int *pixcount, int *offset,
                  int *weight)

{int x, xin, w, nw;
 const unsigned char *urgb;

 for (x = 0; x < out_width; x++)
 {
   xin = offset[x];          /* Where does the pixel come from ? */
   urgb = in_urgb + xin*3;
   w = weight[x];            /* What is the weight for linear interpolation ? */
   nw = 128 - w;
   out_irgb[0] += (urgb[0] * w + urgb[3]*nw)/128;
   out_irgb[1] += (urgb[1] * w + urgb[4]*nw)/128;
   out_irgb[2] += (urgb[2] * w + urgb[5]*nw)/128;
   out_irgb += 3;
   (pixcount[x])++;
 }
}


static void
msc_row_scale_down (const unsigned char *in_urgb, int in_width,
                    int *out_irgb, int *pixcount, int *offset)

{int x, xout;
 register int *irgb;

 for (x = 0; x < in_width; x++)
 {
   xout = offset[x];           /* Where is the output pixel ? */
   irgb = out_irgb + xout*3;
   *(irgb++) += *(in_urgb++);  /* Sum up RGB-values for output pixel */
   *(irgb++) += *(in_urgb++);
   *(irgb++) += *(in_urgb++);
   (pixcount[xout])++;         /* How many pixels have been added ? */
 }
}


static void
msc_row_normalize (register const int *irgb, register unsigned char *rgb_out,
                   int width, int *pixcount)

{register int n;

 while (width-- > 0)
 {
   n = *(pixcount++);
   *(rgb_out++) = (unsigned char)(*(irgb++) / n);
   *(rgb_out++) = (unsigned char)(*(irgb++) / n);
   *(rgb_out++) = (unsigned char)(*(irgb++) / n);
 }
}


static void
msc_columns_interpolate (unsigned char *rgb0, unsigned char *rgb1,
                         int width)
{register unsigned char *rgb;
 register int v0, v1;
 register int *weight;
 int k, nrows, bpr;

 bpr = width*3;
 nrows = (rgb1 - rgb0)/bpr + 1;  /* Number of rows including top and bottom */
 if (nrows <= 2) return;         /* Nothing to do ? */
                                 /* Only the rows inbetween are interesting */
 weight = malloc ((nrows-2) * sizeof (int));
 if (weight == NULL) return;

 for (k = 1; k < nrows-1; k++)
   weight[k-1] = 128 - (k * 128) / (nrows-1);

 nrows -= 2;

 while (width-- > 0)   /* for all columns */
 {                     /* Work on red component */
   rgb = rgb0 + bpr;    v0 = *(rgb0++);    v1 = *(rgb1++);
   for (k = 0; k < nrows; k++)
   {
     *rgb = (unsigned char )(((v0-v1)*weight[k] + 128*v1)/128);
     rgb += bpr;
   }
                       /* Work on green component */
   rgb = rgb0 + bpr;    v0 = *(rgb0++);    v1 = *(rgb1++);
   for (k = 0; k < nrows; k++)
   {
     *rgb = (unsigned char )(((v0-v1)*weight[k] + 128*v1)/128);
     rgb += bpr;
   }
                       /* Work on blue component */
   rgb = rgb0 + bpr;    v0 = *(rgb0++);    v1 = *(rgb1++);
   for (k = 0; k < nrows; k++)
   {
     *rgb = (unsigned char )(((v0-v1)*weight[k] + 128*v1)/128);
     rgb += bpr;
   }
 }
 free (weight);
}


static int
msc_image_scale (MOSAIC_IMAGE *img_out, const MOSAIC_IMAGE *img_in,
                 int width, int height)

{register int *rowoffset = NULL;
 register unsigned char *rgb_in, *rgb_out;
 int *pixcount = NULL, *irgb = NULL, *weight = NULL;
 int dx, dy, yout, next_yout, last_yout;
 int row_scaleup;
 int retval = -1;

 dx = img_in->width - width;
 dy = img_in->height - height;

 if ((dx == 0) && (dy == 0))
 {
   memcpy (img_out->rgb, img_in->rgb, width*3*height);
   return 0;
 }

 row_scaleup = (width > img_in->width);

 pixcount = malloc (width * sizeof (int));
 irgb = malloc (width * 3 * sizeof (int));
 if ((irgb == NULL) || (pixcount == NULL)) goto cleanup;

 if (row_scaleup)
 {
   rowoffset = malloc ((width+1) * sizeof (int));
   weight = malloc (width * sizeof (int));
   if ((rowoffset == NULL) || (weight == NULL)) goto cleanup;
   for (dx = 0; dx < width; dx++)
   {
     rowoffset[dx] = (dx * (img_in->width-1)) / (width - 1);
     if (rowoffset[dx] < img_in->width-1)
       weight[dx] = 128 - ((dx * (img_in->width-1) * 128) / (width-1)) % 128;
     else       /* We must not access elements beyond the edge */
       weight[dx] = 128;
   }
 }
 else
 { /* Mapping table x-source to x-destination */
   rowoffset = malloc (img_in->width * sizeof (int));
   if (rowoffset == NULL) goto cleanup;
   for (dx = 0; dx < img_in->width; dx++)
     rowoffset[dx] = (dx * (width-1)) / (img_in->width-1);
 }

 img_out->width = width;
 img_out->height = height;
 rgb_in = img_in->rgb;

 last_yout = yout = 0;
 memset (irgb, 0, sizeof (int) * width * 3);
 memset (pixcount, 0, sizeof (int) * width);

 for (dy = 0; dy < img_in->height; dy++)
 {
   if (row_scaleup)
     msc_row_scale_up (rgb_in, irgb, width, pixcount, rowoffset, weight);
   else
     msc_row_scale_down (rgb_in, img_in->width, irgb, pixcount, rowoffset);

   /* Are we about to skip to the next output y-line ? */
   next_yout = ((dy+1) * (height-1)) / (img_in->height-1);
   if ((next_yout != yout) || (dy+1 == img_in->height))
   {
     rgb_out = img_out->rgb + yout * width * 3;
     msc_row_normalize (irgb, rgb_out, width, pixcount);
     if (yout-last_yout > 1)   /* Did we skip lines ? */
       msc_columns_interpolate (img_out->rgb + last_yout * width * 3,
                                rgb_out, width);
     last_yout = yout;
     yout = next_yout;
     if (yout >= height) yout = height-1;
     memset (irgb, 0, sizeof (int) * width * 3);
     memset (pixcount, 0, sizeof (int) * width);
   }
   rgb_in += img_in->width*3;
 }
 retval = 0;

cleanup:
 if (irgb) free (irgb);
 if (pixcount) free (pixcount);
 if (rowoffset) free (rowoffset);
 if (weight) free (weight);
 return retval;
}

#endif

#ifdef MSC_STAND_ALONE
static void
msc_image_paste (MOSAIC_IMAGE *img_out, const MOSAIC_IMAGE *img_in,
                 int xpos, int ypos)

{unsigned char *dst;
 const unsigned char *src;
 int k, len;

 if (xpos >= img_out->width) return;
 len = img_in->width;
 if (xpos + len >= img_out->width) len = img_out->width - xpos;

 src = img_in->rgb;
 dst = img_out->rgb + (ypos*img_out->width + xpos)*3;
 for (k = 0; k < img_in->height; k++)
 {
   if (ypos+k>= img_out->height) break;
   memcpy (dst, src, len*3);
   src += img_in->width*3;
   dst += img_out->width*3;
 }
}
#endif


static void
msc_image_crop (MOSAIC_IMAGE *img_out,
                const MOSAIC_IMAGE *img_in,
                int xpos, int ypos, int width, int height)

{unsigned char *src, *dst;
 int k;

 /* It could be possible that we have to crop parts outside of an image. */
 /* We could try to set those parts to 0. But then the cropped part */
 /* would be very dark and a corresponding tile difficult to find. */
 /* Just leave that image content as it is. */
 img_out->width = width;
 img_out->height = height;

 dst = img_out->rgb;
 src = img_in->rgb + (ypos*img_in->width + xpos)*3;
 for (k = 0; k < height; k++)
 {
   if (ypos+k >= img_in->height) break;
   if (xpos+width <= img_in->width)
     memcpy (dst, src, width*3);
   else
     memcpy (dst, src, (img_in->width-xpos)*3);
   dst += width*3;
   src += img_in->width*3;
 }
}


/* Search the best fitting image within an archive that is to replace
 * a single tile.
 *   mia    (I): info about all available archive images
 *   mia_hsv(I): HSV-info for all archive images
 *   cinfo  (I): info about input tile for which the best replacement
 *               has to be found
 *   cinfo_hsv (I): HSV-info about input tile
 *   format (I): image format to use
 *   exclude_grey (I): flag to exclude grey images from search
 *   enable_v_mirror (I): flag to enable mirroring of archive images
 *   reuse (I): Reuse images
 *   used (I/O): flags for each archive image if already used
 *   best_fit (O): Options for best fitting archive image
 * Return value:
 *   Index of best fitting image if found, -1 otherwise.
 *
 * If reuse is 0 and enable_v_mirror is 1, an image can be used twice:
 * mirrored and un-mirrored.
 */
static int
msc_archive_search_best_fit (const MOSAIC_ARCHIVE *mia,
                             const MOSAIC_COLOR_INFO *mia_hsv,
                             const MOSAIC_COLOR_INFO *cinfo,
                             const MOSAIC_COLOR_INFO *cinfo_hsv,
                             MOSAIC_FORMAT format,
                             int exclude_grey,
                             int enable_v_mirror,
                             int reuse,
                             const char *used,
                             MOSAIC_BEST_FIT *best_fit)

{double distance, best_distance = 1.0e99;
 int k, best_index = -1, best_v_mirror = 0;
 MOSAIC_FILE *mia_mf;
 MOSAIC_COLOR_INFO cinfo_mirror;

 memset (best_fit, 0, sizeof (*best_fit));
 best_fit->index = -1;

 for (k = 0, mia_mf = &(mia->mf[0]); k < mia->size; k++, mia_mf++)
 {
   if (mia_mf->cinfo.grey && exclude_grey) continue;
   if (format != msc_image_format (mia_mf->image.width, mia_mf->image.height))
     continue;

   /* Image already used ? */
   if (used)
   {int used_unmirrored = (used[k] & MOSAIC_IMAGE_USED_UNMIRRORED);
    int used_mirrored = (used[k] & MOSAIC_IMAGE_USED_V_MIRRORED);

     if (used_unmirrored && used_mirrored) continue;
     if ((!reuse) && (used_mirrored || used_unmirrored)) continue;
   }

   /* First search pass: used is specified. In that case drop images */
   /* which have large distance in average color. */
   /* But dont do that if reuse of images is not allowed. */
   if (used && (reuse != 0))
   {int dr, dg, db;
     dr = ((int)cinfo->r) - ((int)mia_mf->cinfo.r); if (dr < 0) dr = -dr;
     dg = ((int)cinfo->g) - ((int)mia_mf->cinfo.g); if (dg < 0) dg = -dg;
     db = ((int)cinfo->b) - ((int)mia_mf->cinfo.b); if (db < 0) db = -db;
     if (   (dr > MAX_RGB_DIST) || (dg > MAX_RGB_DIST) || (db > MAX_RGB_DIST)
         || (sqrt ((double)(dr*dr + dg*dg + db*db)) > MAX_RGB_DIST)) continue;
   }

   /* Test unmirrorred image if no use array specified or if image */
   /* has not been already used unmirrored. */
   if ((used == NULL) || !(used[k] & MOSAIC_IMAGE_USED_UNMIRRORED))
   {
     distance = msc_rgb_color_distance(cinfo, &(mia_mf->cinfo));
     if (distance < best_distance)
     {
       best_index = k;
       best_v_mirror = 0;
       best_distance = distance;
     }
   }
   if (!enable_v_mirror) continue;

   /* Now try the mirrored image */
   if ((used == NULL) || !(used[k] & MOSAIC_IMAGE_USED_V_MIRRORED))
   {
     msc_color_info_v_mirror (&(mia_mf->cinfo), &cinfo_mirror);
     distance = msc_rgb_color_distance (cinfo, &cinfo_mirror);
     if (distance < best_distance)
     {
       best_index = k;
       best_v_mirror = 1;
       best_distance = distance;
     }
   }
 }
 if (best_index < 0) return -1;

 best_fit->index = best_index;
 best_fit->v_mirror = best_v_mirror;

 return best_index;
}


/* Search the best fitting images of an archive that are to replace
 * the tiles of an image.
 *   mia        (I): info about all available archive images
 *   num_tiles  (I): number of tiles that are to replace
 *   tile_info  (I/O): info for each tile (will be randomly swapped)
 *   format     (I): format of the archive images to use
 *   exclude_grey (I): flag to exclude grey images for replacement
 *   enable_v_mirror (I): flag to enable vertical mirroring
 *   enable_random   (I): flag to enable random sequence for tile replace
 *   reuse_image     (I): flag to enable multiple use of an image
 *   best_fit   (O): for each tile options of archive image to replace tile
 */
static void
msc_mosaic_search_best_fit (const MOSAIC_ARCHIVE *mia,
                            int num_tiles, TILE_INFO *tile_info,
                            MOSAIC_FORMAT format, int exclude_grey,
                            int enable_v_mirror, int enable_random,
                            int reuse_image,
                            MOSAIC_BEST_FIT *best_fit)
{int k, bf;
 char *used = NULL;
 TILE_INFO ti_tmp;
 MOSAIC_COLOR_INFO *mia_hsv = NULL;
 MOSAIC_COLOR_INFO *tile_hsv = NULL;

 memset (best_fit, 0, sizeof (*best_fit)*num_tiles);
 used = (char *)calloc (mia->size, sizeof (char));
 if (used == NULL) return;

 /* Randomly swap tiles */
 if (enable_random)
 {
   for (k = 0; k < num_tiles; k++)
   {
     bf = msc_rand (num_tiles);
     if (bf == k) continue;
     memcpy (&ti_tmp, tile_info+k, sizeof (*tile_info));
     memcpy (tile_info+k, tile_info+bf, sizeof (*tile_info));
     memcpy (tile_info+bf, &ti_tmp, sizeof (*tile_info));
   }
 }
 for (k = 0; k < num_tiles; k++)
 {
   if (msc_progress_function)
     (*msc_progress_function) ((double)k/(double)num_tiles/2.0);

   /* Search which archive file fits best and is not used already */
   bf = msc_archive_search_best_fit (mia, mia_hsv,
                                     &(tile_info[k].cinfo), tile_hsv+k,
                                     format,
                                     exclude_grey, enable_v_mirror,
                                     reuse_image, used, best_fit+k);

   /* Nothing found ? Find one which is already used. */
   if ((bf < 0) && reuse_image)
   {
     bf = msc_archive_search_best_fit (mia, mia_hsv,
                                       &(tile_info[k].cinfo), tile_hsv+k,
                                       format,
                                       exclude_grey, enable_v_mirror,
                                       reuse_image, NULL, best_fit+k);
   }

   /* Mark it as used */
   if (bf >= 0)
     used[bf] |= best_fit[k].v_mirror ? MOSAIC_IMAGE_USED_V_MIRRORED :
                                        MOSAIC_IMAGE_USED_UNMIRRORED;
 }
 if (used) free (used);
 if (mia_hsv) free (mia_hsv);
 if (tile_hsv) free (tile_hsv);
}


/*** Functions to handle image class ***/

/* Initialize a class image */
void
mscic_init (MOSAIC_IMAGE_CLASS *this, void *image,
            void (*img_size_get)(const struct mosaic_ic_image_s *this,
                                 int *width, int *height),
            int (*img_size_set)(struct mosaic_ic_image_s *this,
                                int width, int height),
            void (*img_crop)(const struct mosaic_ic_image_s *this,
                             MOSAIC_IMAGE *img_out,
                             int xpos, int ypos, int width, int height),
            void (*img_paste)(struct mosaic_ic_image_s *this,
                              int xpos, int ypos, const MOSAIC_IMAGE *img))

{
 memset (this, 0, sizeof (*this));
 this->image = image;
 this->img_size_get = img_size_get;
 this->img_size_set = img_size_set;
 this->img_crop = img_crop;
 this->img_paste = img_paste;
}


#ifdef MSC_STAND_ALONE

/* Get extent of a class image */
static void
mscic_size_get (const MOSAIC_IMAGE_CLASS *this, int *width, int *height)

{const MOSAIC_IMAGE *img = (const MOSAIC_IMAGE *)(this->image);

 *width = img->width;
 *height = img->height;
}


/* Set extent of a class image. This must also create what is necessary */
/* to store data to the image. */
static int
mscic_size_set (MOSAIC_IMAGE_CLASS *this, int width, int height)

{MOSAIC_IMAGE *img = (MOSAIC_IMAGE *)(this->image);

 img->width = width;
 img->height = height;
 img->rgb = malloc (width*height*3);
 if (img->rgb == NULL) return -1;
 return 0;
}


/* Get a part of a class image. The part to retrieve may be partly outside */
/* of the class image. The part outside should not be initialized. */
/* Data is stored in the mosaic image. */
static void
mscic_crop (const MOSAIC_IMAGE_CLASS *this,
            MOSAIC_IMAGE *img_out,
            int xpos, int ypos, int width, int height)

{const MOSAIC_IMAGE *img = (const MOSAIC_IMAGE *)(this->image);

 msc_image_crop (img_out, img, xpos, ypos, width, height);
}


/* Paste a mosaic image into a class image. Parts of the mosaic image */
/* could be outside the class image. */
static void
mscic_paste (MOSAIC_IMAGE_CLASS *this,
             int xpos, int ypos,
             const MOSAIC_IMAGE *img)

{MOSAIC_IMAGE *this_img = (MOSAIC_IMAGE *)(this->image);

 msc_image_paste (this_img, img, xpos, ypos);
}

#endif

static int
mscic_paste_scaled (MOSAIC_IMAGE_CLASS *img_out,
                    int xpos, int ypos, int width, int height,
                    const MOSAIC_IMAGE *img_in)

{int dx, dy;

 dx = img_in->width - width;
 dy = img_in->height - height;

 /* Is the image a little bit larger than the output size ? Just paste */
 if ((dx >= 0) && (dx <= 5) && (dy >= 0) && (dy <= 5))
 {
   (*(img_out->img_paste))(img_out, xpos, ypos, img_in);
 }
 else /* The size of the image is much different. */
 {    /* Crop out a part with the aspect ratio of the destination size */
      /* to prevent different scaling in x/y. We prefer this, but */
      /* it will change the color properties. */
  int crop_x, crop_y, crop_dx, crop_dy;
  MOSAIC_IMAGE *cropped, *scaled;

   crop_dx = img_in->width;
   crop_dy = (crop_dx * height) / width;
   if (crop_dy > img_in->height)
   {
     crop_dy = img_in->height;
     crop_dx = (crop_dy * width) / height;
   }
   crop_x = (img_in->width - crop_dx) / 2;
   crop_y = (img_in->height - crop_dy) / 2;

   cropped = msc_image_alloc (crop_dx, crop_dy);
   if (cropped == NULL) return -1;
   scaled = msc_image_alloc (width, height);
   if (scaled == NULL)
   {
     msc_image_free (cropped);
     return -1;
   }
   msc_image_crop (cropped, img_in, crop_x, crop_y, crop_dx, crop_dy);
   msc_image_scale (scaled, cropped, width, height);
   (*(img_out->img_paste))(img_out, xpos, ypos, scaled);
   msc_image_free (scaled);
   msc_image_free (cropped);
 }
 return 0;
}


/* Create a photo mosaic as a class image. Use a class image as input. */
/* ntiles is a rough estimate for how many tiles should be used. */
/* If out_width/out_height is 0, the size of the output class image is a */
/* multiple of the average size in the image archive. */
int
mscic_mosaic_create (const MOSAIC_IMAGE_CLASS *img_in,
                     const char *archive_name,
                     int ntiles, int out_width, int out_height,
                     int keep_aspect,
                     int exclude_grey,
                     int enable_v_mirror,
                     int enable_random,
                     int reuse_image,
                     MOSAIC_IMAGE_CLASS *img_out)

{int tile_in_xpos, tile_in_ypos;
 double tile_in_width_d, tile_in_height_d;
 int tile_in_width_i, tile_in_height_i;
 int tile_out_xpos, tile_out_ypos;
 double tile_out_width_d, tile_out_height_d;
 int tile_out_width_i, tile_out_height_i, tile_count;
 int max_tiles, num_tiles;
 int in_width, in_height, avrg_width, avrg_height;
 int ev, nx, ny, xc, yc, retval = -1;
 int n_rows, n_columns;
 double aspect_in, aspect_lib;
 MOSAIC_ARCHIVE *mia;
 MOSAIC_IMAGE tile_img_in, lib_img_in;
 MOSAIC_FORMAT format;
 TILE_INFO *ti, *tile_info = NULL;
 char name[MSC_MAX_PATH], buf[MSC_MAX_PATH];
 char dirname[MSC_MAX_PATH], basename[MSC_MAX_PATH];
 unsigned char *tile_img_in_rgb = NULL;
 unsigned char *lib_img_in_rgb = NULL;
 char *printlist;
 MOSAIC_BEST_FIT *best_fit = NULL;

 msc_dirname_get (archive_name, dirname);
 msc_basename_get (archive_name, basename);
 mia = msc_archive_read (dirname, basename);
 if (mia == NULL)
 {
   sprintf (buf, "mscic_mosaic_create: cant read archive %s (%s)",
            archive_name, msc_err_str);
   msc_message (buf);
   goto cleanup;
 }

 memset (&tile_img_in, 0, sizeof (tile_img_in));
 memset (&lib_img_in, 0, sizeof (lib_img_in));

 lib_img_in_rgb = lib_img_in.rgb = malloc (MOSAIC_MAX_SIZE*MOSAIC_MAX_SIZE*3);
 if (lib_img_in.rgb == NULL)
 {
   msc_message ("mscic_mosaic_create: Not enough memory");
   goto cleanup;
 }

 /* Check which image formats have at most in the archive */
 format = MOSAIC_PORTRAIT;
 if (mia->format[MOSAIC_LANDSCAPE].n > mia->format[format].n)
   format = MOSAIC_LANDSCAPE;
 if (mia->format[MOSAIC_SQUARE].n > mia->format[format].n)
   format = MOSAIC_SQUARE;
 avrg_width = mia->format[format].avrg_width;
 avrg_height = mia->format[format].avrg_height;

 /* Get size of input image */
 (*(img_in->img_size_get))(img_in, &in_width, &in_height);

 if ((out_width <= 0) || (out_height <= 0))
   out_width = out_height = 0;
 else if (keep_aspect)
 {int nwidth, nheight;

   nwidth = out_width;
   nheight = (nwidth * in_height) / in_width;
   if (nheight > out_height)
   {
     nheight = out_height;
     nwidth = (nheight * in_width) / in_height;
   }
   out_width = nwidth;
   out_height = nheight;
 }

 /* Aspect ratio of input image and archive files (average) */
 aspect_in = ((double)in_width)/((double)in_height);
 aspect_lib = ((double)avrg_width)/((double)avrg_height);

 /* Get number of tiles in x-direction */
 nx = (int)(sqrt (ntiles * aspect_in / aspect_lib) + 0.5);
 if (nx < 2) nx = 2;
 ny = (int)(nx*aspect_lib/aspect_in + 0.5); /* This is only a rough value. */

 max_tiles = (nx+2)*(ny+2);
 tile_info = (TILE_INFO *)malloc (max_tiles * sizeof (TILE_INFO));
 if (tile_info == NULL)
 {
   msc_message ("mscic_mosaic_create: Not enough memory for tile info file");
   goto cleanup;
 }

 best_fit = (MOSAIC_BEST_FIT *)malloc (max_tiles * sizeof (MOSAIC_BEST_FIT));
 if (best_fit == NULL)
 {
   msc_message ("mscic_mosaic_create: Not enough memory for best fit");
   goto cleanup;
 }

 /* Get size of tiles. They should match the aspect ratio of the archive */
 tile_in_width_d = ((double)in_width) / ((double)nx);
 tile_in_height_d = tile_in_width_d / aspect_lib;

 /* The integer values keep the same. Only size of the output tiles varies */
 tile_in_width_i = (int)(tile_in_width_d+0.5);
 tile_in_height_i = (int)(tile_in_height_d+0.5);

 /* No output size specified ? Try to use unscaled archive files */
 if ((out_width <= 0) || (out_height <= 0))
 {
   out_width = nx * avrg_width;
   out_height = (int)(out_width / aspect_in + 0.5);
 }
 tile_out_width_d = tile_in_width_d/((double)in_width)*((double)out_width);
 tile_out_height_d = tile_in_height_d/((double)in_height)*((double)out_height);

 /* Set size of output image */
 if ((*(img_out->img_size_set))(img_out, out_width, out_height) != 0)
 {
   msc_message ("mscic_mosaic_create: Not enough memory for output file");
   goto cleanup;
 }

 tile_img_in.rgb = malloc ((int)((tile_in_width_d+1.0)*(tile_in_height_d+1)*3));
 tile_img_in_rgb = tile_img_in.rgb;
 if (tile_img_in.rgb == NULL)
 {
   strcpy (msc_err_str, "mscic_mosaic_create: not enough memory");
   goto cleanup;
 }

 /* Evaluate all infos for all tiles of the image */
 num_tiles = 0;
 ti = tile_info;
 n_rows = n_columns = 0;

 for (yc = 0; ; yc++)
 {
   tile_in_ypos = (int)(yc * tile_in_height_d);
   if (tile_in_ypos >= in_height) break;

   if (yc > n_rows) n_rows = yc;

   tile_out_ypos = (int)(yc * tile_out_height_d);
   tile_out_height_i = ((int)((yc+1) * tile_out_height_d)) - tile_out_ypos;

   for (xc = 0; ; xc++)
   {
     if (num_tiles >= max_tiles) continue;

     tile_in_xpos = (int)(xc * tile_in_width_d);
     if (tile_in_xpos >= in_width) break;

     if (xc > n_columns) n_columns = xc;

     tile_out_xpos = (int)(xc * tile_out_width_d);
     tile_out_width_i = ((int)((xc+1) * tile_out_width_d)) - tile_out_xpos;

     /* Crop the part we want to replace */
     (*(img_in->img_crop))(img_in, &tile_img_in, tile_in_xpos, tile_in_ypos,
                           tile_in_width_i, tile_in_height_i);

     /* Get the color information for that part */
     msc_color_info_eval (&tile_img_in, &(ti->cinfo));

     /* Save values */
     ti->in_xpos = tile_in_xpos;
     ti->in_ypos = tile_in_ypos;
     ti->in_width = tile_in_width_i;
     ti->in_height = tile_in_height_i;
     ti->out_xpos = tile_out_xpos;
     ti->out_ypos = tile_out_ypos;
     ti->out_width = tile_out_width_i;
     ti->out_height = tile_out_height_i;

     num_tiles++;
     ti++;
   }
 }

 n_rows++;
 n_columns++;

 msc_mosaic_search_best_fit (mia, num_tiles, tile_info, format, exclude_grey,
                             enable_v_mirror, enable_random, reuse_image,
                             best_fit);

 printlist = getenv ("PMSC_PRINT_LIST");
 if (printlist && (*printlist == '1'))
   msc_mosaic_print_list (n_rows, n_columns, num_tiles, best_fit, mia);

 /* Replace all tiles */
 for (tile_count = 0, ti = tile_info;
      tile_count < num_tiles;
      tile_count++, ti++)
 {
   if (msc_progress_function)
     (*msc_progress_function) ((double)tile_count/(double)num_tiles/2.0+0.5);

   if (best_fit[tile_count].index < 0) /* Replace by part of original image */
   {MOSAIC_IMAGE repl;

     memset (&repl, 0, sizeof (repl));
     repl.rgb = malloc (ti->in_width * ti->in_height * 3);
     if (repl.rgb != NULL)
     {
       (*(img_in->img_crop))(img_in, &repl, ti->in_xpos, ti->in_ypos,
                   ti->in_width, ti->in_height);
       ev = mscic_paste_scaled (img_out, ti->out_xpos, ti->out_ypos,
                                ti->out_width, ti->out_height,
                                &repl);
       free (repl.rgb);
     }
     continue;
   }

   /* Read the best fitting archive file */
   sprintf (name, "%s/%s/%s", mia->directory,
            mia->mf[best_fit[tile_count].index].subdir,
            mia->mf[best_fit[tile_count].index].name);

   ev = msc_jpg_image_read (name, &lib_img_in);
   if (ev != 0) ev = msc_ppm_image_read (name, &lib_img_in);

   if (ev != 0)   /* We did not get the image ? */
   {
     sprintf (buf, "mscic_mosaic_create: Error reading mosaic image (%s)",
              msc_err_str);
     msc_message (buf);
   }
   else   /* We got the image. Paste the scaled image to the output */
   {
     if (best_fit[tile_count].v_mirror) /* Mirror image if requested */
       msc_image_v_mirror (&lib_img_in);

     ev = mscic_paste_scaled (img_out, ti->out_xpos, ti->out_ypos,
                              ti->out_width, ti->out_height,
                              &lib_img_in);
     if (ev != 0)
       msc_message ("mscic_mosaic_create: Error paste scaled image");
   }
 }
 if (msc_progress_function)
   (*msc_progress_function) (1.0);

 retval = 0;

cleanup:
 if (lib_img_in_rgb) free (lib_img_in_rgb);
 if (tile_img_in_rgb) free (tile_img_in_rgb);
 if (best_fit) free (best_fit);
 if (tile_info) free (tile_info);
 if (mia) msc_archive_struct_free (mia);

 return retval;
}

typedef struct
{ const char *name;
  int count;
} TILE_LIST;

/* Print list of tiles */
static void
msc_mosaic_print_list (int n_rows, int n_columns, int num_tiles,
                       const MOSAIC_BEST_FIT *best_fit,
                       const MOSAIC_ARCHIVE *mia)
{int k, j, n_names = 0;
 int row, column;
 TILE_LIST *tile_list = (TILE_LIST *)calloc (num_tiles, sizeof (TILE_LIST));

 if (!tile_list)
 {
   fprintf (stderr, "Not enough memory\n");
   return;
 }

 printf ("\nNumber of tiles: %d (%d rows x %d columns)\n", num_tiles,
         n_rows, n_columns);
 if (n_rows * n_columns != num_tiles)
   printf ("Oops. Number of tiles do not match rows x columns\n");

 row = column = 0;
 printf ("\nList of tiles from top left to bottom right:\n");
 for (k = 0; k < num_tiles; k++)
 {
   const char *tile_name = mia->mf[best_fit[k].index].name;

   if ((k % n_columns) == 0)
     printf ("\nRow %d:\n  ", ++row), column = 0;
   else
     column++;

   printf ("%s ", tile_name);
   if ((column % 4) == 3) printf ("\n  ");

   for (j = 0; j < n_names; j++) /* Search the name in the list */
   {
     if (strcmp (tile_name, tile_list[j].name) == 0)
     {
       (tile_list[j].count)++;
       break;
     }
   }
   if (j >= n_names)  /* Not found ? Add new tile */
   {
     tile_list[n_names].name = tile_name;
     tile_list[n_names++].count = 1;
   }
 }
 printf ("\n\nUsed tiles and count:\n");
 for (j = 0; j < n_names; j++)
   printf ("%s: %d\n", tile_list[j].name, tile_list[j].count);

 free (tile_list);
}

#ifdef MSC_ADD_MAIN

static void
usage (char *prog)

{
 fprintf (stderr, "Usage: %s archive <directory> <archivefile>\n", prog);
 fprintf (stderr, "       %s read <directory> <archivefile>\n", prog);
 fprintf (stderr, "       %s mosaic <ppm-file> <directory> <archivefile>\
 <ntiles> <outputfile> [<width> <height>]\n", prog);
 fprintf (stderr, "       %s collect <srcdir> <dstdir>\n", prog);
}

int
main (int argc, char **argv)

{
 if (argc < 3)
 {
   usage (argv[0]);
   exit (1);
 }

 if ((strcmp (argv[1], "archive") == 0) && (argc == 4))
 {MOSAIC_ARCHIVE *ma, *mb;
  int k;

   ma = msc_archive_create (argv[2]);
   if (ma == NULL)
   {
     fprintf (stderr, "Error in msc_archive_create (%s/%s)\n", argv[2],argv[3]);
     fprintf (stderr, "Error string: %s\n", msc_err_str);
     exit (1);
   }

   if (msc_archive_write (ma, argv[2], argv[3]) != 0)
   {
     fprintf (stderr, "Error in msc_archive_write\n");
     exit (1);
   }

   printf ("Entries in created archive: %d\n", ma->size);

   mb = msc_archive_read (argv[2], argv[3]);
   if (mb == NULL)
   {
     fprintf (stderr, "Error in msc_archive_read (%s/%s)\n", argv[2],argv[3]);
     fprintf (stderr, "Error string: %s\n", msc_err_str);
     exit (1);
   }

   printf ("Entries in read archive: %d\n", mb->size);
   for (k = 0; k < 3; k++)
   {
     printf ("%d images of average size %dx%d\n", mb->format[k].n,
             mb->format[k].avrg_width, mb->format[k].avrg_height);
   }

   msc_archive_struct_free (ma);
   msc_archive_struct_free (mb);
 }
 else if ((strcmp (argv[1], "read") == 0) && (argc == 4))
 {MOSAIC_ARCHIVE *mb;
  int k;

   mb = msc_archive_read (argv[2], argv[3]);
   if (mb == NULL)
   {
     fprintf (stderr, "Error in msc_archive_read (%s/%s)\n", argv[2],argv[3]);
     fprintf (stderr, "Error string: %s\n", msc_err_str);
     exit (1);
   }

   printf ("Entries in read archive: %d\n", mb->size);
   for (k = 0; k < 3; k++)
   {
     printf ("%d images of average size %dx%d\n", mb->format[k].n,
             mb->format[k].avrg_width, mb->format[k].avrg_height);
   }
   msc_archive_struct_free (mb);
 }
 else if ((strcmp (argv[1], "collect") == 0) && (argc == 4))
 {int k;
   k = msc_images_collect (argv[2], argv[3], 128, 97, NULL);
   printf ("image collect returned %d\n", k);
 }
 else if ((strcmp (argv[1], "mosaic") == 0) && ((argc == 7) || (argc == 9)))
 {MOSAIC_IMAGE img_in, img_out;
  MOSAIC_IMAGE_CLASS ic_img_in, ic_img_out;
  char archive[512];
  int width = 0, height = 0;

   if (argc == 9)
   {
     width = atoi (argv[7]);
     height = atoi (argv[8]);
   }
   memset (&img_in, 0, sizeof (img_in));
   memset (&img_out, 0, sizeof (img_out));
   if (msc_ppm_image_read (argv[2], &img_in) != 0)
   {
     fprintf (stderr, "Error in msc_ppm_image_read (%s)\n", argv[2]);
     exit (1);
   }
   sprintf (archive, "%s/%s", argv[3], argv[4]);

   mscic_init (&ic_img_in, &img_in, mscic_size_get, mscic_size_set,
               mscic_crop, mscic_paste);
   mscic_init (&ic_img_out, &img_out, mscic_size_get, mscic_size_set,
               mscic_crop, mscic_paste);

   if (mscic_mosaic_create (&ic_img_in, archive, atoi (argv[5]), width, height,
                            1, 1, 1, 1, 1, &ic_img_out) != 0)
   {
     fprintf (stderr, "Error in mscic_mosaic_create %s %s\n", argv[3],argv[4]);
     fprintf (stderr, "Error string: %s\n", msc_err_str);
     exit (1);
   }
   msc_ppm_image_write (argv[6], &img_out);

   if (img_out.rgb) free (img_out.rgb);
   if (img_in.rgb) free (img_in.rgb);
 }
 else
 {
   usage (argv[0]);
   exit (1);
 }
 return 0;
}

#endif
